function B = Sample_slice( B_prev, Defender, Attacker, ppack )
%Samples the next time step given the previous
%Inputs:    B = current time step
%           Defender = defender
%           Attacker = attacker
%Output:    Bnew = next time step

assert( ppack.params.player_type > 0 && ppack.params.player_type < 3 );
% Continuous player
if ( ppack.params.player_type == 1 )
    % No previous, use priors
    if ( isempty(B_prev) )
        %Set priors
        S_prev.V1 = ppack.params.V0;
        D_prev = 0; % dX = 0 @ t=0
        A_prev.p3 = ppack.params.p3;
        A_prev.q3 = 0;  % m = 0 @ t=0
        MD_prev.SD = 0;
        MD_prev.V1 = 0;
        MD_prev.V2 = 0;
        MD_prev.V3 = 0;
        MA_prev.SA = 0;
        MA_prev.V3 = 0;
        MA_prev.V2_prev = 0;
        MA_prev.V3_prev = 0;
        MA_prev.q3 = 0;
        MA_prev.q3_prev = 0;
    else
        [S_prev, OD_prev, OA_prev, MD_prev, MA_prev, D_prev, A_prev] ...
                    = Slice_unpack(B_prev);
    end

    % Main sample sequence
    S = Sample_S( S_prev, D_prev, A_prev, ppack );
    OD = Sample_OD( S );
    MD = Sample_MD( MD_prev, OD, D_prev, ppack.params );
    
    D = Defender_policy( Defender, MD, ppack.params );
    
    OA = Sample_OA( S );
    MA = Sample_MA( MA_prev, OA, A_prev, ppack.params );

    A = Attacker_policy( Attacker, MA, ppack );
    
    %Copy to compact
    B = Slice_pack( S, OD, OA, MD, MA, D, A );
    
% Discrete player
elseif ( ppack.params.player_type == 2 )
    % No previous, use priors
    if ( isempty(B_prev) )
        %Set priors
        S_prev.V1 = ppack.params.V0;
        D_prev = 0; % dX = 0 @ t=0
        A_prev.p3 = ppack.params.p3;
        A_prev.q3 = 0;  % m = 0 @ t=0
        MD_prev.SD = 0;
        MD_prev.V1 = 0;
        MD_prev.V2 = 0;
        MD_prev.V3 = 0;
        MA_prev.SA = 0;
        MA_prev.V3 = 0;
        MA_prev.V3_prev = 0;
        MA_prev.q3 = 0;
        MA_prev.q3_prev = 0;
        MA_prev.EA = 2;
    else
        [S_prev, OD_prev, OA_prev, MD_prev, MA_prev, D_prev, A_prev] ...
                    = Slice_unpack(B_prev);
    end

    % Main sample sequence
    S = Sample_S( S_prev, D_prev, A_prev, ppack );

    OD = Sample_OD( S );
    MD = Sample_MD(MD_prev, OD, D_prev, ppack.params );
    
    if( isfield(Defender, 'useContL0') && Defender.useContL0 && Defender.mode==3)
        D = Defender_policy( Defender, MD, ppack.params );
    else
        MD_d = Discretize_MD( MD, ppack );
        Dindex = Defender_policy_d( Defender, MD_d, ppack.CPTs, ppack.params, ppack.BinsOrSet );
        D = ppack.params.dXset( Dindex );
    end
    
    OA = Sample_OA( S );
    MA = Sample_MA(MA_prev, OA, A_prev, ppack.params );
    
    if( isfield(Attacker, 'useContL0') && Attacker.useContL0 && Attacker.mode==3)
        A = Attacker_policy( Attacker, MA, ppack);
    else
        MA_d = Discretize_MA( MA, ppack );
        Aindex = Attacker_policy_d( Attacker, MA_d, ppack.CPTs, ppack.params, ppack.BinsOrSet );
        A.p3 = Aindex.p3;
        A.q3 = ppack.BinsOrSet.q3Set(Aindex.q3index);
    end
    
    %Copy to compact
    B = Slice_pack( S, OD, OA, MD, MA, D, A );

end
end

function OD = Sample_OD( S )
%Sample OD from S
    OD.V1 = S.V1;
    OD.V2 = S.V2;
    OD.V3 = S.V3;
end

function OA = Sample_OA( S )
%Sample OA from S
    OA.V2 = S.V2;
    OA.V3 = S.V3;
end

function [MD] = Sample_MD(MD_prev, OD, D_prev, params )
% Sample defender's memory @ t
% Inputs: MD@t-1, OD@t, dXindex@t-1
% Output: MD

    MD.V1 = OD.V1;  % V1 @ t-1
    MD.V2 = OD.V2;  % V2 @ t-1
    MD.V3 = OD.V3;  % V3 @ t-1
    MD.D_prev = D_prev; % dX @ t-1
    
    MD.V1_prev = MD_prev.V1;  % V1 @ t-1
    MD.V2_prev = MD_prev.V2;  % V2 @ t-1
    MD.V3_prev = MD_prev.V3;  % V3 @ t-1

    if( params.mem.mode_Def == 1 )
        % Compute the statistic of defender @ t
        MD.SD = params.mem.gamma_Def * MD_prev.SD...
                    + sign(MD.V1-MD.V1_prev).*sign(MD.V3-MD.V3_prev);
    elseif ( params.mem.mode_Def == 2 )
%         abs((MD.V3-MD_prev.V3)-(MD.V2-MD_prev.V2))
        sameBinDiff = abs((MD.V3-MD_prev.V3)-(MD.V2-MD_prev.V2)) < 0.001;
        if (sameBinDiff)
            MD.SD = MD_prev.SD - 1;
        else
            MD.SD = MD_prev.SD + 3;
        end
        MD.SD = max(MD.SD, 1);
        MD.SD = min(MD.SD, 11);    
    end
end

function [MA] = Sample_MA(MA_prev, OA, A_prev, params )
% Sample attacker's memory @ t
% Inputs: MA@t-1, OA@t, A@t-1, global parameters
% Output: MA


    MA.V2 = OA.V2;               % V2_@ t
    MA.V3 = OA.V3;          % V3_@ t
    MA.V3_prev = MA_prev.V3;      % V3_@ t-1
    MA.V3_2prev = MA_prev.V3_prev; % V3_@ t-2 
    MA.q3_prev = A_prev.q3;      %AA @ t-1
    MA.q3_2prev = MA_prev.q3_prev; %AA @ t-2
    
    dV3 = MA.V3_prev - MA.V3_2prev;
    dq3 = MA.q3_prev - MA.q3_2prev;
    
    % Compute the statistic of attacker @ t
    MA.SA = params.mem.gamma_Att * MA_prev.SA...
                + sign(floor((dV3-dq3*params.x2/params.V0)/...
                (params.dV/params.V0)));
end