function [CPT_ScottDefL0] = GenCPT_ScottDefL0(thetaD, params, BinsOrSet,T)
% This function is used to generate the CPT for Scott's defender's L0
% policy
% Inputs: Parameter thetaD representing the defender's characteristics
%         Global parameters
%         Bins for V1, V2, V3
% Output: CPT_ScottDefL0

    % Initialize CPT_ScottDefL0
    CPT_ScottDefL0 = zeros(BinsOrSet.N_dX, BinsOrSet.N_V1, BinsOrSet.N_V2,...
                           BinsOrSet.N_V3, params.N_statisticD );
    
    if ( ~exist('T') )
        T = 30000; % Number of times we repeat for each combination of inputs
    end
    
    if ( params.routeSD2D0 == 1 ) % Don't route memory to thetaD
        N_outerloops = 1;
    elseif ( params.routeSD2D0 == 2 ) % Route memory to thetaD
        N_outerloops = params.N_statisticD;  %for each bin of memory statistic
    else
        assert(false);
    end
    
    for l = 1:N_outerloops
        SD = (l-1)./(params.N_statisticD-1);
        for k = 1:BinsOrSet.N_V3 % for each bin of V3
            % Generate V3 uniformly from the bin
            if (k == 1) % Apply half-Gussian distribution. Only sample the left to the mean 
                mu = BinsOrSet.V2n3bins(2); % mean = upperbound of the lowest bin
                sigma = (BinsOrSet.V2n3bins(3)-BinsOrSet.V2n3bins(2))/sqrt(1-2/pi); % Stdev = range of right neighboring bin divided by a difference factor
                Ran = mu + sigma .* randn(T,1); % Generate full-normally distributed numbers
                V3 = BinsOrSet.V2n3bins(2) - abs(BinsOrSet.V2n3bins(2) - Ran); % Convert full-normal to half-normal
            elseif (k == BinsOrSet.N_V3) % Apply half-Guassian distribution. Only sample the right to the mean. 
                mu = BinsOrSet.V2n3bins(k); % mean = lowerbound of the highest bin
                sigma = (BinsOrSet.V2n3bins(k) - BinsOrSet.V2n3bins(k-1))/sqrt(1-2/pi); % Stdev = range of the left neighboring bin divided by a difference factor
                Ran = mu + sigma .* randn(T,1); % Generate full-normally distributed numbers
                V3 = BinsOrSet.V2n3bins(k) + abs(BinsOrSet.V2n3bins(k) - Ran); % Convert full-normal to half-normal
            else % Sample uniformly in the medium three bins
                V3 = BinsOrSet.V2n3bins(k) + (BinsOrSet.V2n3bins(k+1)...
                    - BinsOrSet.V2n3bins(k)) .* rand(T,1);
            end

            for j = 1:BinsOrSet.N_V2 % for each bin of V3
                % Generate V2 uniformly from the bin
                if (j == 1) % Apply half-Gussian distribution. Only sample the left to the mean 
                    mu = BinsOrSet.V2n3bins(2); % mean = upperbound of the lowest bin
                    sigma = (BinsOrSet.V2n3bins(3)-BinsOrSet.V2n3bins(2))/sqrt(1-2/pi); % Stdev = range of right neighboring bin divided by a difference factor
                    Ran = mu + sigma .* randn(T,1); % Generate full-normally distributed numbers
                    V2 = BinsOrSet.V2n3bins(2) - abs(BinsOrSet.V2n3bins(2) - Ran); % Convert full-normal to half-normal
                elseif (j == BinsOrSet.N_V3) % Apply half-Guassian distribution. Only sample the right to the mean. 
                    mu = BinsOrSet.V2n3bins(j); % mean = lowerbound of the highest bin
                    sigma = (BinsOrSet.V2n3bins(j) - BinsOrSet.V2n3bins(j-1))/sqrt(1-2/pi); % Stdev = range of the left neighboring bin divided by a difference factor
                    Ran = mu + sigma .* randn(T,1); % Generate full-normally distributed numbers
                    V2 = BinsOrSet.V2n3bins(j) + abs(BinsOrSet.V2n3bins(j) - Ran); % Convert full-normal to half-normal
                else % Sample uniformly in the medium three bins
                    V2 = BinsOrSet.V2n3bins(j) + (BinsOrSet.V2n3bins(j+1)...
                        - BinsOrSet.V2n3bins(j)) .* rand(T,1);
                end

                for i = 1:BinsOrSet.N_V1 % for each discrete point of V1
                    V1 = BinsOrSet.V1set(i) * ones(T,1); % Repeat T times
                    x = (V1 - params.V0)*params.N / params.dV; % Back out x 

                    Num = zeros(BinsOrSet.N_dX, 1); % Initialize the tracker

                    % Scott's defender's L0 policy
                    z = ((V2+V3)/2-params.V0)/(params.dV/params.N);
                    PiD_normal =  -sign(z) .* sign( floor( abs(z) )); % In normal operations
                    PiD_safe = -sign(x-1.02); % if defender believes attacker presents at node 3
                    
                    if ( params.routeSD2D0 == 2 )
                        thetaD = SD;  % if route, then override, otherwise use input arg
                    end
                    
                    % Smoothly move between two policy extremes
                    dX = sign((1-thetaD).*PiD_normal + thetaD.*PiD_safe);
                    [~,p] = histc(dX, params.dXset);

                    for s = 1:BinsOrSet.N_dX % for each value of dX
                        Num(s,1) = sum(p==s); % Count the number of times this value occurs
                    end
                    CPT_ScottDefL0(:,i,j,k,l) = Num(:,1)/T; % Conditional probabilities 
                end
            end
        end
    end
end
                
                
                
                
                
                
                