function [Defender] = Train_defender_SARSANN( Attacker, params )
%Learn an RL Defender assuming given an Attacker
%using SARSA and a Neural Network approximator.
%This version uses epoch-wise updating.
%Input: Attacker = attacker
%Output: Defender = learned RL defender
%global params;

fprintf('Sarsa_defender_NN...\n')

params.sarsann = Set_params_SARSANN();

%create a new defender
clear Defender;
Defender.mode = params.policy_mode;     %Put defender into policy mode

%Initialize neural net
N_neurons = 50;    %number of neurons in the hidden layer of the NN
scale = 0.5;  %scales the initial random weights
%Declare a new feedforward (multi-layer perceptron) NN
Defender.thetaD = newff( [0.5,1.5; 0.5,1.5; 0.5,1.5; -2,2; -2,2; -1,1; 
        -5,5; -1,1],[0, 2], [N_neurons], {'tansig' 'purelin'});
Defender.thetaD.IW{1} = scale*rand(N_neurons,8);
Defender.thetaD.LW{2,1} = scale*rand(1,N_neurons);
Defender.thetaD.b{1} = scale*rand(N_neurons,1);
Defender.thetaD.b{2} = scale*rand(1,1);
%configure the regulator
Defender.thetaD.divideFcn = '';
Defender.thetaD.performFcn = 'msereg';
Defender.thetaD.performParam.ratio = 0.1;
%configure learning rate
Defender.thetaD.inputWeights{1}.learnParam.lr = params.alpha;
Defender.thetaD.layerWeights{2,1}.learnParam.lr = params.alpha;
Defender.thetaD.biases{1}.learnParam.lr = params.alpha;
Defender.thetaD.biases{2}.learnParam.lr = params.alpha;

%Initialize tracking variables
Defender.trainReward = zeros(params.nTrainEpsDef,params.nSimSteps);
Defender.trainIO = zeros(10, params.nTrainEpsDef*params.nSimSteps);
Defender.trainErr = zeros(params.nTrainEpsDef, params.nSimSteps);
Defender.trainPerf = zeros(1, params.nTrainEpsDef);
Defender.trainLW = zeros(N_neurons,params.nTrainEpsDef);

%Parameter scheduling
%No scheduling = constant at the desired eps
if ( params.schedule_mode == 0 )
    schedule = params.epsilonD * ones(1,params.nTrainEpsDef);
%3-segment scheduling
elseif( params.schedule_mode == 2)
   %eps-greedy case
   if( params.policy_mode == 7 )
       schedule = Schedule_3_seg(params.nTrainEpsDef,params.schedule_start,params.epsilonD);
   else
       assert(false);   %error
   end
else
    assert(false);  %error
end

trackIndex = 0; %for tracking
for ep = 1:params.nTrainEpsDef
   ep
   
   %a place to collect all the NN training data
   inputs = zeros(8,params.nSimSteps);
   Qtargets = zeros(1,params.nSimSteps);
   
   %deal with epsilon scheduling
   Defender.eps = schedule(ep);
   epsilon = schedule(ep)
   B = Sample_slice([], Defender, Attacker,params);
   
   for t = 1:params.nSimSteps
       %Extract variables at t
       [S_t,OD_t,MD_t,dX_t,OA_t,MA_t,m_t,L_t] = Slice_unpack(B);
       a_t = dX_t;
       phi_t = [OD_t; MD_t(1)]; %"s_t" = [Observation; Memory(3)]
       Qsa_t = sim(Defender.thetaD, [phi_t; a_t]);   %simulate neural net
       
       %Sample the next step
       B = Sample_slice(B, Defender, Attacker,params);
       
       %Extract variables at t+1
       [S_tp1,OD_tp1,MD_tp1,dX_tp1,OA_tp1,MA_tp1,m_tp1,L_tp1] = Slice_unpack(B);
       a_tp1 = dX_tp1;
       phi_tp1 = [OD_tp1; MD_tp1(1)]; %"s_t+1" = [Observation; Memory(3)]
       Qsa_tp1 = sim(Defender.thetaD, [phi_tp1; a_tp1]);   %simulate neural net
       r_tp1 = Defender_reward(B,params);
       
       %SARSA
       if ( params.learn_mode == 1 )
           Qtarget = r_tp1 + params.gamma*Qsa_tp1;  %desired Q-value
       %Q-learning
       elseif ( params.learn_mode == 2 )
           Qtarget = r_tp1 + params.gamma*max_Q(Defender.thetaD,phi_tp1,params);
       else
           assert(false);   %error
       end
       
       inputs(:,t) = [phi_t; a_t];
       Qtargets(t) = Qtarget;
       
       %For tracking
       trackIndex = trackIndex + 1;
       Defender.trainIO(:,trackIndex) = [phi_t; a_t; Qsa_t; Qtarget];
       Defender.trainReward(ep,t) = r_tp1;
   end
   
   %train NN at the end of the episode
   Defender.thetaD.adaptParam.passes = 20;      %adapt faster by making multiple passes
   r = randperm(params.nSimSteps);  %randomize the permutations to learn better
   [Defender.thetaD,Y,E,Pf,Af,tr] = adapt(Defender.thetaD,inputs(:,r),Qtargets(:,r));  %adapt NN
   
   %For tracking
   Defender.trainErr(ep,:) = E;
   Defender.trainPerf(ep) = tr.perf;
   Defender.trainLW(:,ep) = Defender.thetaD.LW{2,1}';
   Defender.thetaD.LW{2,1}';
   assert(max(Defender.thetaD.LW{2,1}) < 1e10 ); %make sure the weights aren't exploding
   
   %display training progress/the mean reward this episode
   mean_reward = mean(Defender.trainReward(ep,:))
end

end

function Qsa_max = max_Q( net, phi,params )
%Returns the max of Qsa over possible actions
%used in Q-learning
%global params;

Qsa = zeros(3,1);
for a_tp1 = [-1,0,1]
    index = a_tp1 + 2;
    Qsa(index) = sim( net, [phi; a_tp1] );
end
Qsa;
Qsa_max = max(Qsa);

end