%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% This is a wrapper class for a power grid system to generize (at least
% somewhat) interactions with power grid models in MatPower format.  This
% assumes that matpower is installed for Matlab (added to the paths).
% Assume $MATPOWER is the directory for Matpower, then you need the
% following directories added to your path.
% $MATPOWER
% $MATPOWER\t
% $MATPOWER\extra
% one way to do this is to modify the startup.m file (in toolbox\local)...
% for example
% addpath XXXX\matpower4.1
% addpath XXXX\matpower4.1\t
% addpath XXXX\matpower4.1\extras
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
classdef PowerGridSystem
    properties (SetAccess = private)
      power_grid; % power_grid model
      results; % results of the simulation   
      attributes; % lookup table for attributes
      gen_map; % lookup table for the generator table
    end % properties

    properties (Constant)
        VOLTAGE_MAGNITUDE   = 'VOLTAGE_MAGNITUDE'; 
        VOLTAGE_ANGLE       = 'VOLTAGE_ANGLE'; 
        REAL_LOAD           = 'REAL_LOAD';    
        REACTIVE_LOAD       = 'REACTIVE_LOAD';
        REAL_GENERATION     = 'REAL_GENERATION';
        REACTIVE_GENERATION = 'REACTIVE_GENERATION';
        REAL_FLOW           = 'REAL_FLOW';
        REACTIVE_FLOW       = 'REACTIVE_FLOW';
        DEFAULT_VALUE       = -9999;
    end % properties
    
   methods
        % Constructor
        function PGS = PowerGridSystem(filename)
            PGS.power_grid = loadcase(filename);
            PGS.gen_map = java.util.HashMap;
            num = size(PGS.power_grid.gen, 1);
            for i = 1:num
                PGS.gen_map.put(java.lang.Integer(PGS.power_grid.gen(i,1)),i);
            end
            
        end % Constructor
        
        % Simulate the power grid system
        function PGS = simulate(PGS)
            options = mpoption;
            options = mpoption(options, 'OUT_ALL', 0); % turn off the output
            options = mpoption(options, 'VERBOSE', 0); % turn off the output
            PGS.results = runpf(PGS.power_grid, options);
        end % simulate

        % get an arbitrary attribute
        function v = getAttribute(PGS, component, attribute)
            define_constants;
            v = PGS.DEFAULT_VALUE;
            if (strcmp(attribute, PGS.VOLTAGE_MAGNITUDE) == 1)
                v = PGS.results.bus(component, VM);
            end
            if (strcmp(attribute, PGS.VOLTAGE_ANGLE) == 1)
                v = PGS.results.bus(component, VA);
            end
            if (strcmp(attribute, PGS.REAL_LOAD) == 1)
                v = PGS.results.bus(component, PD);
            end            
            if (strcmp(attribute, PGS.REACTIVE_LOAD) == 1)
                v = PGS.results.bus(component, QD);
            end            
            if (strcmp(attribute, PGS.REAL_GENERATION) == 1)
                v = PGS.results.gen(PGS.gen_map.get(java.lang.Integer(component)), PG);
            end            
            if (strcmp(attribute, PGS.REACTIVE_GENERATION) == 1)
                v = PGS.results.gen(PGS.gen_map.get(java.lang.Integer(component)), QG);
            end            
            if (strcmp(attribute, PGS.REAL_FLOW) == 1)
                v = PGS.results.branch(component, PF);
            end            
            if (strcmp(attribute, PGS.REACTIVE_FLOW) == 1)
                v = PGS.results.branch(component, QF);
            end            
        end % getAttribute

        % set an arbitrary attribute
        function PGS = setAttribute(PGS, component, attribute, value)
            define_constants;
            if (strcmp(attribute, PGS.VOLTAGE_MAGNITUDE) == 1)
                PGS.power_grid.bus(component, VM) = value;
                PGS.power_grid.gen(PGS.gen_map.get(java.lang.Integer(component)), VG) = value;         
            end
            if (strcmp(attribute, PGS.VOLTAGE_ANGLE) == 1)
                PGS.power_grid.bus(component, VA) = value;
            end
            if (strcmp(attribute, PGS.REAL_LOAD) == 1)
                PGS.power_grid.bus(component, PD) = value;
            end
            if (strcmp(attribute, PGS.REACTIVE_LOAD) == 1)
                PGS.power_grid.bus(component, QD) = value;
            end
            if (strcmp(attribute, PGS.REAL_GENERATION) == 1)
%                disp(PGS.power_grid.gen(PGS.gen_map.get(java.lang.Integer(component)), PG))
 %               disp(value)
                PGS.power_grid.gen(PGS.gen_map.get(java.lang.Integer(component)), PG) = value;
  %              disp(PGS.power_grid.gen(PGS.gen_map.get(java.lang.Integer(component)), PG))
            end
            if (strcmp(attribute, PGS.REACTIVE_GENERATION) == 1)
                %disp(PGS.power_grid.gen(PGS.gen_map.get(java.lang.Integer(component)), QG))
                %disp(value)
                PGS.power_grid.gen(PGS.gen_map.get(java.lang.Integer(component)), QG) = value;
                %disp(PGS.power_grid.gen(PGS.gen_map.get(java.lang.Integer(component)), QG))
            end
        end % setAttribute
        
   end % methods 
end % classdef

