function [player, info] = Train_Jaakkola_c(SIM, playerList, playerNumber, makeplot)
%Learn a RL player assuming given other players
%using continuous Jaakkola algorithm.
%This version uses binned observations and continuous state
%equations for the players
%Input: SIM = simulation object
%       playerList = vector of player objects
%       playerNumber = the position in the playerList that you want to
%       train
%Output: player = learned RL player

fprintf(strcat('Jaakkola Continuous: Player ', num2str(playerNumber), '...'));

params.jaak = Set_params_jaak_c(playerNumber);
info.trainParams = params.jaak;

%Default makeplot to true
if(~exist('makeplot'))
    makeplot = true;
end



%get number of elements in the memory and in the move space of the player
moveDim = size(playerList(playerNumber).moveSet,2); %dimension of move space (1 node, 2node?) for player
for i=1:moveDim
    MoveNumOfElm(i) = numel(playerList(playerNumber).moveSet{i}); %# of possible moves
end
MemNumOfElm = playerList(playerNumber).GetMemNumOfElm;
RLinputNumOfElm = MemNumOfElm + moveDim; %Total number of inputs to RL algorithm: memory plus action
MemNumOfBinsVec = playerList(playerNumber).GetMemNumOfBinsVec();

%initialize parameters
r_av = 0;
K = ones([MemNumOfBinsVec,MoveNumOfElm]);
beta = K; 
Q = info.trainParams.InitialQ*K;
C=K;
info.AvRewPerPlcy = zeros(1,info.trainParams.nTrainEps);
info.AvRewOverall = info.AvRewPerPlcy;
info.pi = cell(1,info.trainParams.nTrainEps);
info.Q = info.pi;
info.K = info.pi;
info.beta = info.pi;
info.C = info.pi;
info.mems = cell(1,info.trainParams.nTrainEps*info.trainParams.NumInnerSteps);

%create a new player
props = playerList(playerNumber).GetProperties();
props.pi = K/prod(MoveNumOfElm);
playerList(playerNumber) = playerList(playerNumber).SetMode( 'Jaakkola_c', props );

%create SIM object
SIM = SIM.SetPlayerList( playerList );

%define the bin edges (if needed) for the player moves
for i=1:moveDim
    MovebinEdges{i} = [-Inf, (playerList(playerNumber).moveSet{i}(1:end-1)+ ...
        playerList(playerNumber).moveSet{i}(2:end))/2, Inf];
end


%for each episode, do
ep = 1;
while ep < info.trainParams.nTrainEps
    ep
    [SIM, mems, moves, rewards] = SIM.Init();
    [SIM, mems, moves, rewards] = SIM.Step(mems, moves);
    
    InnerLoopRewAv = 0;
    for t = 1:info.trainParams.NumInnerSteps
        
        %calculate the reward
        r = rewards{playerNumber};
            
        %discretize the memory and the moves (find bin indexes
        % of the memory and the moves)
        mem_d.vec = num2cell(playerList(playerNumber).GetMemBinInd(mems{playerNumber}));
        for i=1:moveDim
            [~,move_d_temp(i)] = histc(moves{playerNumber}{i}, MovebinEdges{i});
        end
        move_d = num2cell(move_d_temp); %converts move_d_temp into a cell array
                   
        K(mem_d.vec{:},move_d{:})=K(mem_d.vec{:},move_d{:}) + 1;
        beta=info.trainParams.gamma*beta;
        Q=Q+beta*(r-r_av);
        
        Q(mem_d.vec{:},move_d{:}) = Q(mem_d.vec{:},move_d{:}) - beta(mem_d.vec{:},move_d{:})*(r-r_av);
        beta(mem_d.vec{:},move_d{:})=(1-(1/K(mem_d.vec{:},move_d{:})))*beta(mem_d.vec{:},move_d{:})+1/K(mem_d.vec{:},move_d{:});
        Q(mem_d.vec{:},move_d{:})=(1-(1/K(mem_d.vec{:},move_d{:})))*Q(mem_d.vec{:},move_d{:})+beta(mem_d.vec{:},move_d{:})*(r-r_av);
        
        %Update the average reward estimate with the current reward
        r_av_new = r_av + (r-r_av)/(info.trainParams.NumInnerSteps*(ep-1)+t);
        InnerLoopRewAv = InnerLoopRewAv + (r-InnerLoopRewAv)/t;
        
        %Transform Jaakkola message-action matrices M(m,a) for m_t=m,
        %a_t=a using the new average reward estimate
        C = C + beta;
        Q = Q - C*(r_av_new-r_av);
        
        Q(mem_d.vec{:},move_d{:}) = Q(mem_d.vec{:},move_d{:}) + C(mem_d.vec{:},move_d{:})*(r_av_new-r_av);
        C(mem_d.vec{:},move_d{:}) = C(mem_d.vec{:},move_d{:}) - beta(mem_d.vec{:},move_d{:});
        
        C(mem_d.vec{:},move_d{:}) = (1-(1/K(mem_d.vec{:},move_d{:})))*C(mem_d.vec{:},move_d{:})+beta(mem_d.vec{:},move_d{:});
        Q(mem_d.vec{:},move_d{:}) = Q(mem_d.vec{:},move_d{:})-C(mem_d.vec{:},move_d{:})*(r_av_new-r_av);
        
        r_av = r_av_new;
    
        %Sample the next step
        [SIM, mems, moves, rewards] = SIM.Step(mems, moves);
        info.mems{info.trainParams.NumInnerSteps*(ep-1)+t} = mems{playerNumber}.vec;
        
    end
    
    props.pi=(1-info.trainParams.epsln)*props.pi;
    
    Q_temp = Q;
    m = max(max(Q_temp,[],RLinputNumOfElm),[],RLinputNumOfElm-1);
    t = bsxfun(@eq,m,Q_temp); % Identify all multiple maxima along each dim 6
    f = find(t); % Get pointers to all of them
    
    indexx = f;
    
    props.pi(indexx) = ...
        props.pi(indexx) + info.trainParams.epsln*1;
    
    %normz = repmat(sum(props.pi,RLinputNumOfElm), [ones(1,MemNumOfElm), MoveNumOfElm]);
    
    %MoveNumOfElm_cell = num2cell(MoveNumOfElm); %cell array is needed to calculate 
                                                %normz below
    normz = repmat(sum(sum(props.pi,RLinputNumOfElm),RLinputNumOfElm-1), [ones(1,MemNumOfElm), MoveNumOfElm]);
    props.pi = props.pi ./ normz;
    
    % Update player and the SIM object
    playerList(playerNumber) = playerList(playerNumber).SetMode( 'Jaakkola_c', props );
    SIM = SIM.SetPlayerList( playerList );
    
    % Track average overall average reward and per policy average reward
    info.AvRewPerPlcy(ep) = InnerLoopRewAv;
    info.AvRewOverall(ep) = r_av;
    info.pi{ep} = props.pi;
    info.Q{ep} = Q;
    info.K{ep} = K;
    info.beta{ep} = beta;
    info.C{ep} = C;
    
    ep=ep+1;
end
player = playerList(playerNumber);

if makeplot
    figure
    plot(info.AvRewPerPlcy);
    title('Average Reward Per Policy');
    
    figure
    plot(info.AvRewOverall);
    title('Overall Average Reward');
        
end

end
    
