
// *************************************************************************
//	LibNFG: Net-form game library
//
//	Author:		Ritchie Lee
//
//  Copyright 2011 United States Government as represented by the Administrator 
//  of the National Aeronautics and Space Administration.  All Rights Reserved.  
//  Licensed under the Apache License, Version 2.0 (the "License"); you may not 
//  use this file except in compliance with the License.  You may obtain a copy 
//  of the License at http://www.apache.org/licenses/LICENSE-2.0
//  Unless required by applicable law or agreed to in writing, software 
//  distributed under the License is distributed on an "AS IS" BASIS, WITHOUT 
//  WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the 
//  License for the specific language governing permissions and limitations 
//  under the License.
//
//	Revision Log
//
//	Username	Date		Description
//	-------------------------------------------------------
//	rlee		20110328	Created
//	rlee		20110720	Added importance sampling capability
//
// *************************************************************************

#include "alg_levelkrelaxed.h"

#include <boost/lambda/bind.hpp>

using namespace std;
using namespace nfg;
using namespace boost;
using namespace boost::lambda;

// *************************************************************************
// *************************************************************************
alg_LevelKRelaxed::alg_LevelKRelaxed( void )
{
	//	Default values
	m_mode					= D_RELAXED;
	m_M						= 5;	
	m_MP					= 10;
	m_normalize				= true;
	m_response				= BEST;
	m_lambda				= 1;


}
// *************************************************************************
// *************************************************************************
alg_LevelKRelaxed::~alg_LevelKRelaxed ( void )
{
}
// *************************************************************************
// *************************************************************************
nfg_Vecs alg_LevelKRelaxed::Get_parent_values( BayesNet& bnet, 
											  const nfg_OptInst& optInst, 
												const nfg_Vertex& v )
{
	nfg_Vecs paVecs;
	//	Loop through each parent of v
	foreach ( nfg_Vertex vi, inv_adjacent_vertices(v, bnet.Get_graph()) )
	{
		// Make sure it is defined
		bassert ( optInst.at(vi) );

		// add it to the list of parent vecs in order of vertex id
		paVecs.push_back( *optInst.at(vi) );
	}

	return paVecs;
}
// *************************************************************************
// *************************************************************************
void alg_LevelKRelaxed::Set_params( BayesNet& bnet, const LEVELK_ALGO mode, 
								   const int M, const int MP, 
								  const nfg_SampleFuncs& fo_Sample_level0_inst,
								  const nfg_SampleFuncs& fo_Sample_move_inst,
								  const LEVELK_RESPONSE response,
								  const bool normalize, const double lambda )
{
	m_mode		= mode;		// Set mode
	
	//	d-relaxed needs d-separated set list
	if ( m_mode == D_RELAXED )
		m_dSepMaskList	= Get_mask_dsep_all( bnet );
	//	others don't need to know the d-separated set list
	else
		m_dSepMaskList.clear();

	//	Perform checks for valid input
	bassert ( M == M && M > 0 );		// Check to make sure it is finite and positive
	bassert ( MP == MP && MP > 0 );		// Check to make sure it is finite and positive
	bassert ( lambda == lambda && lambda >= 0.0 );	// Check that lambda is finite and positive

	//	Make sure every decision node has a level0 function object associated with it
	for ( nfg_Vertex vi = 0; vi < fo_Sample_level0_inst.size(); vi++ )
	{
		if ( bnet.Get_vertex_type(vi) == DECISION )
			bassert ( fo_Sample_level0_inst.at(vi) );		// Will return false if invalid
	}

	//	Assign to member variables
	m_M							= M;
	m_MP						= MP;
	m_fo_Sample_level0_inst		= fo_Sample_level0_inst;
	m_fo_Sample_move_inst		= fo_Sample_move_inst;
	m_lambda					= lambda;
	m_normalize					= normalize;
	m_response					= response;

	return;
}
// *************************************************************************
// *************************************************************************
void alg_LevelKRelaxed::Set_player_proposal_funcs( BayesNet& bnet, 
										   const nfg_PropFuncInsts& propFuncInsts,
										const nfg_PropProbFuncInsts& propProbFuncInsts )
{
	// Check sizes.  Should be size of bnet
	bassert( propFuncInsts.size() == propProbFuncInsts.size() );
	bassert( propFuncInsts.size() == bnet.Get_num_vertices() );

	m_propFuncInsts			= propFuncInsts;
	m_propProbFuncInsts		= propProbFuncInsts;
}
// *************************************************************************
// *************************************************************************
nfg_WtdInst alg_LevelKRelaxed::LW_forward_sample( BayesNet& bnet, const int K )
{

	//	All players are assumed to be at the same level K
	//	No partial masking, instantiations, or evidence
	nfg_OptWtdInst optWtdInst 
		= LW_forward_sample_detailed ( bnet, nfg_OptInst(), K, nfg_Mask(), nfg_OptWtdInst(),
					nfg_PropFuncInst(), nfg_PropProbFuncInst() );

	return nfg_WtdInst( Convert_OptInst_to_Inst( optWtdInst.m_optInst ), 
				optWtdInst.m_weight );
}
// *************************************************************************
// *************************************************************************
nfg_OptWtdInst alg_LevelKRelaxed::LW_forward_sample_detailed( BayesNet& bnet, 
												const nfg_OptInst& evidence,
												const int K, const nfg_Mask& mask, 
												const nfg_OptWtdInst& partialWtdInst,
												const nfg_PropFuncInst& propFuncInst,
												const nfg_PropProbFuncInst& propProbFuncInst )
{
	//	Forward Sampling Algorithm for CHANCE and DECISION nodes.
	//	Calls LevelK_LW_forward_sample_node_prop on each node to do the distinguishing

	//	Check that K >= 1 and finite
	bassert ( K == K && K >= 0 );

	//	For convenience
	unsigned int N = bnet.Get_num_vertices();

	//	Check that mask is valid, should be size of bnet num vertices
	bassert ( mask.empty() || mask.size() == N );

	// Handle proposal distribution functions
	// Only 2 valid sizes, 0 or num nodes. Check that.
	bassert( propFuncInst.size() == propProbFuncInst.size() );
	bassert( propFuncInst.empty() || propFuncInst.size() == N );

	// Resize to not get a read error
	nfg_PropFuncInst				propFuncInstResized( propFuncInst );
	nfg_PropProbFuncInst			propProbFuncInstResized( propProbFuncInst );
	propFuncInstResized.resize( N );
	propProbFuncInstResized.resize( N );

	nfg_OptWtdInst optWtdInst;
	// Maintain a stack of pointers to the partial insts (1) to track hierarchy, and
	// (2) for importance sampling use
	m_partialInstPtrs.push_back(&optWtdInst);

	//	A valid partial exists.  Start from this.
	if ( partialWtdInst.m_optInst.size() == N )
	{	
		optWtdInst.m_optInst	= partialWtdInst.m_optInst;
		optWtdInst.m_weight	= partialWtdInst.m_weight;
	}
	//	No partial exists, so start fresh
	else if ( partialWtdInst.m_optInst.empty() )
	{
		optWtdInst.m_optInst.resize(N);
		optWtdInst.m_weight = 1.0;
	}
	//	Something is wrong...
	else
	{
		bassert(false);
	}

	foreach ( nfg_Vertex vi, vertices( bnet.Get_graph() ) )
	{	
		//	Sample if this vertex is not being masked
		if ( mask.empty() || mask.at(vi) )
		{
			//	Get list of parent values
			nfg_Vecs paVecs = Get_parent_values ( bnet, optWtdInst.m_optInst, vi ) ;

			nfg_WtdVec wVec;
			//	If node has a dno value and dno function evaluates to true
			if ( bnet.Get_vertex_dno(vi) && 
					bnet.Get_dno_func(vi)( optWtdInst.m_optInst ) )
			{
				wVec.m_vec.clear();	// Mark as dno (disabled)
				wVec.m_weight = 1;
			}
			//	If it is not dno, sample the node
			else
			{
				// If prop function exists, preload proposal distributions and do sample
				if ( propFuncInstResized.at(vi) && propProbFuncInstResized.at(vi) )
				{
					nfg_PropFuncDirect propFunc( bind(propFuncInstResized.at(vi),*m_partialInstPtrs.front()) );
					nfg_PropProbFuncDirect propProbFunc( bind(propProbFuncInstResized.at(vi),_1,*m_partialInstPtrs.front()) );
					// Do the sampling
					wVec = LW_forward_sample_node_prop ( bnet, evidence, vi, paVecs, K,
								propFunc,
								propProbFunc );
				}
				// If proposal doesn't exist, pass in empty
				else
				{
					// Do the sampling
					wVec = LW_forward_sample_node_prop ( bnet, evidence, vi, paVecs, K,
								nfg_PropFuncDirect(),
								nfg_PropProbFuncDirect() );
				}
			}
			
			//	Slot it into the right place in the instantiation
			optWtdInst.m_optInst.at(vi) = wVec.m_vec;
			//	Update the weight of the instantiation
			optWtdInst.m_weight *= wVec.m_weight;
		}
	}

	// Completed inst, so remove this from stack
	m_partialInstPtrs.pop_back();

	return optWtdInst;
}
// *************************************************************************
// *************************************************************************
nfg_WtdVec alg_LevelKRelaxed::LW_forward_sample_node_prop ( BayesNet& bnet, 
														   const nfg_OptInst& evidence,
															const nfg_Vertex id, 
															const nfg_Vecs& paVecs,
															const int K,
															const nfg_PropFuncDirect& propFunc,
															const nfg_PropProbFuncDirect& propProbFunc )
{
	nfg_WtdVec wVec;
	wVec.m_weight = 1.0;
	
	//	if not in evidence set, sample the node
	if ( evidence.empty() || !evidence.at(id) )
	{
		//	Sample if Chance node
		if ( bnet.Get_vertex_type(id) == CHANCE )
		{
			// If proposal distribution is given, use that
			if ( propFunc )
			{
				wVec.m_vec	= propFunc();
				double Qprob	= propProbFunc(wVec.m_vec);
				
				// Check that prob of proposal is a valid probability
				bassert( Qprob == Qprob && Qprob > 0 && Qprob <= 1.0 );

				// Correction weight
				wVec.m_weight = bnet.Get_cpd(id)->Get_probability_prop( wVec.m_vec, paVecs ) / Qprob;
			}
			// If not, use natural distribution
			else
				wVec.m_vec	= bnet.Get_cpd(id)->Sample_node( paVecs );
		}
		//	Run Level K (d-)relaxed strategy if Decision node
		else if ( bnet.Get_vertex_type(id) == DECISION )
		{
			if ( m_mode == D_RELAXED )
				wVec.m_vec	= LevelK_d_relaxed( bnet, id, paVecs, K );
			else if ( m_mode == RELAXED )
				wVec.m_vec	= LevelK_relaxed( bnet, id, paVecs, K );
			else
				bassert(false); //	Other modes not supported
			
		}
		//	Other types not supported
		else
		{
			bassert ( false );
		}
	}
	//	if the node is in the evidence set, set the likelihood weight
	else
	{
		//	if in evidence set, assign value to node and correct for it using weight
		wVec.m_vec		= *evidence.at(id);
		wVec.m_weight	*= bnet.Get_cpd(id)->Get_probability_prop( *evidence.at(id), paVecs );
	}

	// Check sampled vector size against bnet declarations
	if(!(wVec.m_vec.size() == bnet.Get_vertex_size(id)))
	{
		cout << "ERROR! Size mismatch. Declared size is: " << bnet.Get_vertex_size(id) 
					<< endl <<endl;
		bassert(false);
	}

	// Make sure weight is finite and nonnegative
	bassert(wVec.m_weight == wVec.m_weight && wVec.m_weight >= 0);

	return wVec;
}
// *************************************************************************
// *************************************************************************
nfg_Vec alg_LevelKRelaxed::LevelK_relaxed ( BayesNet& bnet, const nfg_Vertex v, 
									const nfg_Vecs& paVecs,
									const int K )
{
	//	For convenience,
	int N	= bnet.Get_num_vertices();

	// Check again to make sure size is correct
	bassert( m_propFuncInsts.size() == m_propProbFuncInsts.size() );
	// Resize so that each entry exists (and is empty) when reading off
	if ( m_propFuncInsts.empty() )
	{
		m_propFuncInsts.resize( N );
		m_propProbFuncInsts.resize( N );
	}

	nfg_Vec outVec;

	//	Level 0 is the end condition, return a sample from L0 distribution
	if ( K == 0 )
	{
		outVec = m_fo_Sample_level0_inst.at(v)( paVecs );
	}
	else
	{
		//	If K >= 1,
		//	Run level K relaxed algorithm
		nfg_OptInst evidence( N );

		//	The parents are evidence for the player
		int pa_i = 0;
		foreach( nfg_Vertex vi, inv_adjacent_vertices(v, bnet.Get_graph()) )
		{
			evidence.at(vi) = paVecs.at(pa_i++);
		}
		
		//	LevelK relaxed algorithm
		//	Step 1: IID sample M own strategies from movespace
		nfg_Vecs ownStrategies( m_M );
		foreach( nfg_Vec& move, ownStrategies )
		{
			move = m_fo_Sample_move_inst.at(v)( paVecs );	// sample satisficing distr.
		}
		//	Remove duplicates
		Sort_and_remove_duplicates( ownStrategies );


		//	Step 2: IID sample MP everything else assuming parents are evidence.
		nfg_OptWtdInsts optWtdInst( m_MP );
		nfg_Vec utilVec;
		foreach ( nfg_Vec move, ownStrategies )
		{
			//	take the candidate move as evidence in the sampling
			evidence.at(v) = move;

			//	Start an accumulator for keeping track of weighted average
			nfg_AccumWtdSum utilAcc;
			foreach ( nfg_OptWtdInst& owi, optWtdInst )
			{
				owi = LW_forward_sample_detailed ( bnet, evidence, K-1, nfg_Mask(), nfg_OptWtdInst(),
							m_propFuncInsts.at(v), m_propProbFuncInsts.at(v) );	// complete the sample

				double util = bnet.Get_cpd(v)->Get_util(
					Convert_OptInst_to_Inst(owi.m_optInst) );	//	compute utility of the sampled inst
				utilAcc ( util, weight = owi.m_weight );			//	accumulate the utility values
			}

			//	utility vector tracks the estimated expected values
			//	m_normalize is true => use normalized estimator
			//	m_normalize is false => use unnormalized estimator
			double wsum			= weighted_sum(utilAcc);
            double sweights		= sum_of_weights(utilAcc);
			if ( m_normalize ) 
			{
                if ( sweights == 0 ) 
				{ 
					bassert( wsum == 0 ); 
					utilVec.push_back(0); 
				}
                else 
					utilVec.push_back( wsum / sweights );
            } 
			else 
				utilVec.push_back( wsum );
		}

		//	Step 3: Choose the move according to response policy
		//	Best response, return the argmax move
		if( m_response == BEST )
		{
			int max_ind = Argmax( utilVec );
			outVec = ownStrategies.at( max_ind );
		}
		//	Logit response, do a weighted sample according to exponential
		else if ( m_response == LOGIT )
		{
			nfg_Vec stratVec = Exponentiate( utilVec, m_lambda );	// take exponential
			stratVec = Normalize( stratVec );						// normalize to 1
			int ind = Weighted_sample ( Convert_pdf_to_cdf(stratVec) );	// pick index randomly according to pdf
			outVec = ownStrategies.at( ind );							// play the strategy corresponding to that index
		}
		else
		{
			//	Other responses not supported
			bassert(false);
		}
	}
	
	return outVec;
}
// *************************************************************************
// *************************************************************************
nfg_Vec alg_LevelKRelaxed::LevelK_d_relaxed ( BayesNet& bnet, 
											 const nfg_Vertex v, 
											const nfg_Vecs& paVecs,
											const int K )
{
	//	For convenience, copy commonly-used values over
	int N	= bnet.Get_num_vertices();

	// Check again to make sure size is correct
	bassert( m_propFuncInsts.size() == m_propProbFuncInsts.size() );
	// Resize so that each entry exists (and is empty) when reading off
	if ( m_propFuncInsts.empty() )
	{
		m_propFuncInsts.resize( N );
		m_propProbFuncInsts.resize( N );
	}

	nfg_Vec outVec;

	//	Level 0 is the end condition, return a sample from L0 distribution
	if ( K == 0 )
	{
		outVec = m_fo_Sample_level0_inst.at(v)( paVecs );
	}
	else
	{
		//	If K >= 1,
		//	Run level K d-relaxed algorithm
		nfg_OptInst evidence( N );

		//	The parents are evidence for the player
		int pa_i = 0;
		foreach( nfg_Vertex vi, inv_adjacent_vertices(v, bnet.Get_graph()) )
		{
			evidence.at(vi) = paVecs.at(pa_i++);
		}
		
		//	LevelK d-relaxed algorithm
		//	Step 1: IID sample M own strategies from movespace
		nfg_Vecs ownStrategies( m_M );
		foreach( nfg_Vec& move, ownStrategies )
		{
			move = m_fo_Sample_move_inst.at(v)( paVecs );	// sample satisficing distr.
		}
		//	Remove duplicates
		Sort_and_remove_duplicates( ownStrategies );


		//	Step 2: IID sample MP non-successors assuming parents are evidence.
		//	For d-relaxed, set mask
		nfg_Mask dSepMask_v			= m_dSepMaskList.at(v);
		nfg_Mask dSepMaskNeg_v		= Get_neg_mask(dSepMask_v);

		nfg_OptWtdInsts dsepWtdInst( m_MP );
		foreach ( nfg_OptWtdInst& owi, dsepWtdInst )
		{
			owi = LW_forward_sample_detailed ( bnet, evidence, K-1, dSepMask_v, 
													nfg_OptWtdInst(), m_propFuncInsts.at(v),
													m_propProbFuncInsts.at(v) );
		}

		//	Step 3: Evaluate the outcome for each combination of move and dsep
		nfg_Vec utilVec;
		foreach ( nfg_Vec move, ownStrategies )
		{
			//	take the candidate move as evidence in the sampling
			evidence.at(v) = move;

			//	Start an accumulator for keeping track of weighted average
			nfg_AccumWtdSum utilAcc;
			foreach ( nfg_OptWtdInst wi, dsepWtdInst )
			{
				nfg_OptWtdInst optWtdInst = LW_forward_sample_detailed ( bnet, evidence, K-1, 
														dSepMaskNeg_v, wi,
														m_propFuncInsts.at(v),
														m_propProbFuncInsts.at(v) );

				double util = bnet.Get_cpd(v)->Get_util(
					Convert_OptInst_to_Inst(optWtdInst.m_optInst) );	//	compute utility of the sampled inst
				utilAcc ( util, weight = optWtdInst.m_weight );				//	accumulate the utility values
			}
			//	utility vector tracks the estimated expected values
			//	m_normalize is true => use normalized estimator
			//	m_normalize is false => use unnormalized estimator
            double wsum			= weighted_sum(utilAcc);
            double sweights		= sum_of_weights(utilAcc);
			if ( m_normalize ) 
			{
                bassert( sweights );		// Complain if denominator is zero
				utilVec.push_back( wsum / sweights );
            } 
			else 
				utilVec.push_back( wsum );
		}

		//	Step 4: Choose the move according to response policy
		//	Best response, return the argmax move
		if( m_response == BEST )
		{
			int max_ind = Argmax( utilVec );		// index from argmax
			outVec = ownStrategies.at( max_ind );	// play corresponding strategy
		}
		//	Logit response, do a weighted sample according to exponential
		else if ( m_response == LOGIT )
		{
			nfg_Vec stratVec = Exponentiate( utilVec, m_lambda );	// take exponential
			stratVec = Normalize( stratVec );						// normalize to 1
			int ind = Weighted_sample ( Convert_pdf_to_cdf(stratVec) );	// pick index randomly according to pdf
			outVec = ownStrategies.at( ind );							// play corresponding strategy
		}
		else
		{
			//	Other responses not supported
			bassert(false);
		}
	}
	
	return outVec;
}
// *************************************************************************
// *************************************************************************
nfg_Mask alg_LevelKRelaxed::Get_neg_mask ( const nfg_Mask& mask )
{
	nfg_Mask negMask;
	foreach ( bool val, mask )
	{
		//	negate each value and push it onto a new list
		negMask.push_back(!val);
	}

	return negMask;
}
// *************************************************************************
// *************************************************************************
nfg_Masks alg_LevelKRelaxed::Get_mask_dsep_all ( BayesNet& bnet )
{
	nfg_Masks maskList(bnet.Get_num_vertices());
	
	//	For each Decision node in the net, make a mask
	//	and add it to the right position in the list
	foreach ( nfg_Vertex vi, vertices(bnet.Get_graph()) )
	{
		if ( bnet.Get_vertex_type(vi) == DECISION )
			maskList.at(vi) = Get_mask_dsep(bnet, vi);
	}

	return maskList;
}
// *************************************************************************
// *************************************************************************
nfg_Mask alg_LevelKRelaxed::Get_mask_dsep ( BayesNet& bnet, const nfg_Vertex& v )
{
	//	Assumes that parents are only evidence in determining the d-separated set

	nfg_Mask mask(bnet.Get_num_vertices());

	//	Go through all nodes in the net and consider each relative to v
	foreach ( nfg_Vertex vi, vertices(bnet.Get_graph()) )
	{
		//	v is not a non-succ (i.e. false)
		if ( vi == v )
		{
			mask.at(vi) = false;
		}
		//	other nodes depend on their parents
		//	If any one of their parents is a succ (i.e. false), then 
		//	they are also a succ (i.e. false)
		else
		{
			mask.at(vi) = true;
			//	Loop through all parents
			foreach ( nfg_Vertex vpa, inv_adjacent_vertices(vi, bnet.Get_graph()) )
			{
				mask.at(vi) = mask.at(vi) && mask.at(vpa);
			}
		}
	}

	return mask;
}
// *************************************************************************
// *************************************************************************
nfg_Masks alg_LevelKRelaxed::Get_mask_succ_all ( BayesNet& bnet )
{
	nfg_Masks maskList(bnet.Get_num_vertices());
	
	//	For each Decision node in the net, make a mask
	//	and add it to the right position in the list
	foreach ( nfg_Vertex vi, vertices(bnet.Get_graph()) )
	{
		if ( bnet.Get_vertex_type(vi) == DECISION )
			maskList.at(vi) = Get_mask_succ(bnet, vi);
	}

	return maskList;
}
// *************************************************************************
// *************************************************************************
nfg_Mask alg_LevelKRelaxed::Get_mask_succ ( BayesNet& bnet, const nfg_Vertex& v )
{
	nfg_Mask mask(bnet.Get_num_vertices());

	//	Go through all nodes in the net and consider each relative to v
	foreach ( nfg_Vertex vi, vertices(bnet.Get_graph()) )
	{
		//	If any one of its parents is v or a succ (i.e. true), then 
		//	they are also a succ (i.e. true)
		mask.at(vi) = false;
		//	Loop through all parents
		foreach ( nfg_Vertex vpa, inv_adjacent_vertices(vi, bnet.Get_graph()) )
		{
			if ( vpa == v || mask.at(vpa) )
			{
				mask.at(vi) = true;
			}
		}
	}

	return mask;
}




