
// *************************************************************************
//	LibNFG: Net-form game library
//
//	Author:		Ritchie Lee
//
//  Copyright 2011 United States Government as represented by the Administrator 
//  of the National Aeronautics and Space Administration.  All Rights Reserved.  
//  Licensed under the Apache License, Version 2.0 (the "License"); you may not 
//  use this file except in compliance with the License.  You may obtain a copy 
//  of the License at http://www.apache.org/licenses/LICENSE-2.0
//  Unless required by applicable law or agreed to in writing, software 
//  distributed under the License is distributed on an "AS IS" BASIS, WITHOUT 
//  WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the 
//  License for the specific language governing permissions and limitations 
//  under the License.
//
//	Revision Log
//
//	Username	Date		Description
//	-------------------------------------------------------
//	rlee		20101025	Created
//	rlee		20110302	Added Level K capabilities
//
// *************************************************************************

#include "alg_montecarlo.h"

#include <boost/function.hpp>
#include "if_matlab.h"

using namespace std;
using namespace nfg;
using namespace boost;

// *************************************************************************
// *************************************************************************
alg_MonteCarlo::alg_MonteCarlo( void )
{
}
// *************************************************************************
// *************************************************************************
alg_MonteCarlo::~alg_MonteCarlo ( void )
{
}
// *************************************************************************
// *************************************************************************
nfg_Inst alg_MonteCarlo::Forward_sample( BayesNet& bnet )
{
	//	Forward Sampling Algorithm 12.1 in Koller's Probabilistic Graphical Models

	//	This function is meant for chance nodes only, check that this is true.
	foreach ( nfg_Vertex vi, vertices( bnet.Get_graph() ) )
		bassert ( bnet.Get_vertex_type(vi) == CHANCE );

	nfg_OptInst optInst( bnet.Get_num_vertices() );

	//	Loop through each vertex in graph
	foreach ( nfg_Vertex vi, vertices( bnet.Get_graph() ) )
	{		
		//	Get list of parent values
		nfg_Vecs paVecs = Get_parent_values ( bnet, optInst, vi ) ;
		
		//	Sample node and save results
		optInst.at( vi ) = bnet.Get_cpd(vi)->Sample_node( paVecs );
	}

	return Convert_OptInst_to_Inst( bnet, optInst );
}
// *************************************************************************
// *************************************************************************
nfg_Inst alg_MonteCarlo::Convert_OptInst_to_Inst ( BayesNet& bnet, const nfg_OptInst& optInst )
{
	// Container for complete weighted instantiation
	nfg_Inst inst(bnet.Get_num_vertices());

	// Make sure it is a complete instantiation
	foreach( nfg_Vertex vi, vertices(bnet.Get_graph()) )
	{
		bassert ( optInst.at(vi) );		// make sure that this exists
		inst.at(vi) = *optInst.at(vi);	// copy it over to the output one
	}

	return inst;
}
// *************************************************************************
// *************************************************************************
nfg_Vecs alg_MonteCarlo::Get_parent_values( BayesNet& bnet, const nfg_OptInst& optInst, 
												const nfg_Vertex& v )
{
	nfg_Vecs paVecs;
	//	Loop through each parent of v
	foreach ( nfg_Vertex vi, inv_adjacent_vertices(v, bnet.Get_graph()) )
	{
		// Make sure it is defined
		bassert ( optInst.at(vi) );

		//	add it to the list of parent vecs in order of vertex id
		paVecs.push_back( *optInst.at(vi) );
	}

	return paVecs;
}
// *************************************************************************
// *************************************************************************
nfg_WtdInst alg_MonteCarlo::LW_forward_sample ( BayesNet& bnet, const nfg_OptInst& evidence )
{
	//	Likelihood-weighted particle generation Algorithm 12.2 
	//		in Koller's Probabilistic Graphical Models
	//	This function is for all chance node bnets only. Check.
	foreach ( nfg_Vertex vi, vertices(bnet.Get_graph())  )
	{
		bassert ( bnet.Get_vertex_type(vi) == CHANCE );
	}

	nfg_OptInst optInst ( bnet.Get_num_vertices() );
	double w = 1.0;

	foreach ( nfg_Vertex vi, vertices(bnet.Get_graph())  )
	{
		//	Get list of parent values
		nfg_Vecs paVecs = Get_parent_values ( bnet, optInst, vi ) ;
		
		//	If not in evidence set
		if ( !evidence.at(vi) )
		{
			//	sample from distribution
			optInst.at(vi)	= bnet.Get_cpd(vi)->Sample_node( paVecs );
		}
		//	if in evidence set,
		else
		{
			//	assign value to node and correct weight
			optInst.at(vi)	= *evidence.at(vi);
			w *= bnet.Get_cpd(vi)->Get_probability_prop( *evidence.at(vi), paVecs );
		}
	}

	// Assemble the output object and return
	return nfg_WtdInst(Convert_OptInst_to_Inst(bnet,optInst),w);
}
// *************************************************************************
// *************************************************************************
double alg_MonteCarlo::LW_estimator ( BayesNet& bnet,
									const int numSamples,
									const nfg_OptInst& evidence,
									const nfg_UtilFunc& func,
									const bool normalized  )
{
	//	Check that it is finite and bigger than 0
	bassert ( numSamples == numSamples && numSamples > 0 );

	//	Start an accumulator
	nfg_AccumWtdSum acc;

	//	Loop through specified number of samples
	for ( int i = 0; i < numSamples; i++ )
	{
		nfg_WtdInst	wInst = LW_forward_sample ( bnet, evidence );	// use likelihood weighted sampling
		acc ( func(wInst.m_inst), weight = wInst.m_weight );		//	use accumulator, keep stats, discard sample to save memomry																
	}

	//	return normalized or unnormalized estimated result depending on setting
	return normalized ? weighted_sum(acc) / sum_of_weights(acc) : weighted_sum(acc);
}

