
// *************************************************************************
//	LibNFG: Net-form game library
//
//	Author:		Ritchie Lee
//
//  Copyright 2011 United States Government as represented by the Administrator 
//  of the National Aeronautics and Space Administration.  All Rights Reserved.  
//  Licensed under the Apache License, Version 2.0 (the "License"); you may not 
//  use this file except in compliance with the License.  You may obtain a copy 
//  of the License at http://www.apache.org/licenses/LICENSE-2.0
//  Unless required by applicable law or agreed to in writing, software 
//  distributed under the License is distributed on an "AS IS" BASIS, WITHOUT 
//  WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the 
//  License for the specific language governing permissions and limitations 
//  under the License.
//
//	Revision Log
//
//	Username	Date		Description
//	-------------------------------------------------------
//	rlee		20101025	Created
//	rlee		20110302	Added LevelK functions
//
// *************************************************************************

#ifndef INCLUDED_ALG_MONTECARLO_H
#define INCLUDED_ALG_MONTECARLO_H

#include "bayesnet.h"

#include <boost/function.hpp>

using namespace boost;

//	Use this class for bnets that comprise only CHANCE nodes
class alg_MonteCarlo
{
public:

	/**	Constructor */
	alg_MonteCarlo( void );

	/**	Destructor */
	virtual ~ alg_MonteCarlo();

	/// Variables

	///	Methods
	/**
	 * Get_parent_values: Get values of parents for a node v
	 *	bnet - nfg object
	 *	inst - could be an instantiation or partial instantiation containing the values of the parents of v
	 *	v - vertex id
	 *	returns: a vector (size num parents of v) of vectors containing the values of the parents
	 *		ordered by ascending parent vertex id.
	 */
	nfg_Vecs		Get_parent_values( BayesNet& bnet, const nfg_OptInst& optInst, 
										const nfg_Vertex& v );
	/**
	 * Convert_OptInst_to_Inst: Convert an nfg_OptInst object to nfg_Inst
	 *		making sure that all values are valid (i.e., complete).  
	 *		vector.empty() entries are "does not occur" flags
	 *	bnet - nfg object
	 *	optInst - optional weighted instantiation object to be checked and copied
	 *	returns: a complete instantiation which represents a sample of the net.
	 */
	nfg_Inst		Convert_OptInst_to_Inst( BayesNet& bnet, 
											const nfg_OptInst& optInst );
	/**
	 * Forward_sample: Returns a top-level sample of the net-form game.  
	 *		Assumes empty evidence set. CHANCE nodes only.
	 *	bnet - nfg object
	 *	returns: instantiation which represents a sample of the net.
	 */
	nfg_Inst		Forward_sample( BayesNet& bnet );
	/**
	 * LW_forward_sample: Returns a top-level sample of the net-form game.  
	 *		Uses likelihood weighting to take care of evidence. CHANCE nodes only.
	 *	bnet - nfg object
	 *	evidence - Size of the BayesNet. Any observed node is specified respective vertex id slot.
	 *	returns: weighted instantiation which represents a sample of the net.
	 */
	nfg_WtdInst		LW_forward_sample ( BayesNet& bnet, const nfg_OptInst& evidence );
	/**
	 * LW_estimator: Collects LW samples from LW_forward_sample and estimates a function value.  
	 *	bnet - nfg object
	 *	numSamples - number of samples of the net to use in the calculation
	 *	evidence - instantiation-sized. Any observed node is specified respective vertex id slot.
	 *	func - function object that takes an instantiation of the net and returns a double
	 *	normalized - true = use normalized estimator, false = unnormalized estimator
	 *	returns: estimated expected value.
	 */
	double			LW_estimator ( BayesNet& bnet, const int numSamples,  
							const nfg_OptInst& evidence, const nfg_UtilFunc& func,
							const bool normalized );

private:

	/// Variables


};

#endif

