
// *************************************************************************
//	LibNFG: Net-form game library
//
//	Author:		Ritchie Lee
//
//  Copyright 2011 United States Government as represented by the Administrator 
//  of the National Aeronautics and Space Administration.  All Rights Reserved.  
//  Licensed under the Apache License, Version 2.0 (the "License"); you may not 
//  use this file except in compliance with the License.  You may obtain a copy 
//  of the License at http://www.apache.org/licenses/LICENSE-2.0
//  Unless required by applicable law or agreed to in writing, software 
//  distributed under the License is distributed on an "AS IS" BASIS, WITHOUT 
//  WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the 
//  License for the specific language governing permissions and limitations 
//  under the License.
//
//	Revision Log
//
//	Username	Date		Description
//	-------------------------------------------------------
//	rlee		20110526	Created
//
// *************************************************************************

#include "cpd_bernoulli.h"

#include <boost/random.hpp>
#include <boost/math/distributions/bernoulli.hpp>

using namespace nfg;
using namespace boost::math;

// *************************************************************************
// *************************************************************************
cpd_Bernoulli::cpd_Bernoulli( const nfg_Vec& probSuccs )
{
	// Make sure the parameters are not empty
	bassert ( !probSuccs.empty() );
	
	//	Check to make sure all variances are positive and finite
	foreach ( double p, probSuccs )
	{
		bassert ( p == p );
		bassert ( p >= 0 && p <= 1.0 );
	}

	m_probs		= probSuccs;
}
// *************************************************************************
// *************************************************************************
cpd_Bernoulli::~cpd_Bernoulli ( void )
{
}
// *************************************************************************
// *************************************************************************
nfg_Vec cpd_Bernoulli::Sample_node ( const nfg_Vecs& paVecs )
{
	//	assert no input vectors. Used as a prior only
	bassert ( paVecs.empty() );

	//	Use helper function call to generate a gaussian sample centered at 
	//	parent vector and with pre-specified member variance.
	nfg_Vec outVec;
	foreach( double p, m_probs )
	{
		outVec.push_back( Rand_bernoulli(p) );
	}

	return outVec;
}
// *************************************************************************
// *************************************************************************
double cpd_Bernoulli::Get_probability( const nfg_Vec& outVec, const nfg_Vecs& paVecs )
{
	//	assert no input vectors.  Used as a prior only.
	bassert ( paVecs.empty() );

	//	assert sizes of meanRates and output are the same
	bassert ( m_probs.size() == outVec.size() );
	
	double prob = 1.0;

	for ( unsigned int i = 0; i < m_probs.size(); i++ )
	{
		//	Use helper function call to get the probability density at that point
		bernoulli distr( m_probs.at(i) );
		//	Multiply all the probabilities together in one long chain
		prob *= pdf(distr, outVec.at(i));
	}
	
	bassert ( prob == prob );	// Make sure it is finite
	bassert ( prob >= 0 && prob <= 1 );	// Make sure it is between 0 and 1

	return prob;
}
// *************************************************************************
// *************************************************************************
void cpd_Bernoulli::Unit_test ( void )
{
	//	Set up variance vector
	cpd_Bernoulli bern( nfg_Vec(2,0.5) );

	//	Set up parents vector list
	nfg_Vecs paVecs;

	//	Test sample_node()
	nfg_Vec outVec = bern.Sample_node( paVecs );
	Print_vec(outVec);

	//	Test Get_prob()
	cout << "Prob: " << bern.Get_probability_prop( outVec,  paVecs ) << endl;

	cout << "Unit test completed." << endl << endl;
}

