
// *************************************************************************
//	LibNFG: Net-form game library
//
//	Author:		Ritchie Lee
//
//  Copyright 2011 United States Government as represented by the Administrator 
//  of the National Aeronautics and Space Administration.  All Rights Reserved.  
//  Licensed under the Apache License, Version 2.0 (the "License"); you may not 
//  use this file except in compliance with the License.  You may obtain a copy 
//  of the License at http://www.apache.org/licenses/LICENSE-2.0
//  Unless required by applicable law or agreed to in writing, software 
//  distributed under the License is distributed on an "AS IS" BASIS, WITHOUT 
//  WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the 
//  License for the specific language governing permissions and limitations 
//  under the License.
//
//	Revision Log
//
//	Username	Date		Description
//	-------------------------------------------------------
//	rlee		20101025	Created
//
// *************************************************************************

#include "cpd_poisson.h"

#include <boost/random.hpp>
#include <boost/math/distributions/poisson.hpp>

using namespace nfg;
using namespace boost::math;

// *************************************************************************
// *************************************************************************
cpd_Poisson::cpd_Poisson( const nfg_Vec& meanRates )
{
	// Acts as a prior distribution set by a fixed meanRate. i.e., no parent vecs
	
	// Make sure the parameters are not empty
	bassert ( !meanRates.empty() );

	foreach ( double meanRate, meanRates )
		bassert( meanRate > 0 );		// math restriction

	// Assign to 
	m_lambdas = meanRates;
}
// *************************************************************************
// *************************************************************************
cpd_Poisson::~cpd_Poisson ( void )
{
}
// *************************************************************************
// *************************************************************************
nfg_Vec cpd_Poisson::Sample_node ( const nfg_Vecs& paVecs )
{
	//	assert no input vectors
	bassert ( paVecs.empty() );

	//	Use helper function call to generate a gaussian sample centered at 
	//	parent vector and with pre-specified member variance.
	nfg_Vec outVec;
	foreach( double lambda, m_lambdas )
	{
		outVec.push_back( Rand_poisson(lambda) );
	}

	return outVec;
}
// *************************************************************************
// *************************************************************************
double cpd_Poisson::Get_probability ( const nfg_Vec& outVec, const nfg_Vecs& paVecs )
{
	//	assert no input vectors
	bassert ( paVecs.empty() );

	//	assert sizes of meanRates and output are the same
	bassert ( m_lambdas.size() == outVec.size() );
	
	double prob = 1.0;

	for ( unsigned int i = 0; i < m_lambdas.size(); i++ )
	{
		//	Use helper function call to get the probability density at that point
		poisson distr( m_lambdas.at(i) );
		//	Multiply all the probabilities together in one long chain
		prob *= pdf(distr, outVec.at(i));
	}
	
	bassert ( prob == prob );	// Make sure it is finite
	bassert ( prob >= 0 && prob <= 1 );	// Make sure it is between 0 and 1

	return prob;
}
// *************************************************************************
// *************************************************************************
void cpd_Poisson::Unit_test ( void )
{
	//	Set up variance vector
	cpd_Poisson pois( nfg_Vec(2,5) );

	//	Set up parents vector list
	nfg_Vecs paVecs;

	//	Test sample_node()
	nfg_Vec outVec = pois.Sample_node( paVecs );
	Print_vec(outVec);

	//	Test Get_prob()
	cout << "Prob: " << pois.Get_probability( outVec,  paVecs ) << endl;

	cout << "Unit test completed." << endl << endl;
}

