
// *************************************************************************
//	LibNFG: Net-form game library
//
//	Author:		Ritchie Lee
//
//  Copyright 2011 United States Government as represented by the Administrator 
//  of the National Aeronautics and Space Administration.  All Rights Reserved.  
//  Licensed under the Apache License, Version 2.0 (the "License"); you may not 
//  use this file except in compliance with the License.  You may obtain a copy 
//  of the License at http://www.apache.org/licenses/LICENSE-2.0
//  Unless required by applicable law or agreed to in writing, software 
//  distributed under the License is distributed on an "AS IS" BASIS, WITHOUT 
//  WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the 
//  License for the specific language governing permissions and limitations 
//  under the License.
//
//	Revision Log
//
//	Username	Date		Description
//	-------------------------------------------------------
//	rlee		20110516	Created
//
// *************************************************************************

#include "cpd_reroute.h"

using namespace nfg;

// *************************************************************************
// *************************************************************************
cpd_Reroute::cpd_Reroute( const nfg_Dim& paIndex )
{
	// Save to data member
	m_paIndex		= paIndex;
}
// *************************************************************************
// *************************************************************************
cpd_Reroute::~cpd_Reroute ( void )
{
}
// *************************************************************************
// *************************************************************************
nfg_Vec cpd_Reroute::Sample_node ( const nfg_Vecs& paVecs )
{
	//	assert only 1 input vector
	bassert ( paVecs.size() == 1 );

	nfg_Vec outVec;

	// Extract the desired components from parent vector
	// using indices in m_paIndex
	foreach( unsigned int ind, m_paIndex )
		outVec.push_back( paVecs.at(0).at(ind) );

	return outVec;
}
// *************************************************************************
// *************************************************************************
double cpd_Reroute::Get_probability ( const nfg_Vec& outVec, const nfg_Vecs& paVecs )
{
	//	make sure only 1 input vector
	bassert ( paVecs.size() == 1 );

	//	make sure sizes of output and stored indices are the same
	bassert ( outVec.size() == m_paIndex.size() );
	
	double prob = 0.0; 

	// This is a deterministic node.  If the output matches evaluated, 
	// then prob is 1
	if ( outVec == Sample_node( paVecs ) )
		prob = 1.0;
	// Otherwise prob is 0
	else
		prob = 0.0;

	return prob;
}
// *************************************************************************
// *************************************************************************
void cpd_Reroute::Unit_test ( void )
{
	//	Set up the routing
	nfg_Dim route;
	route.push_back(2);	// indices of the parent vec
	route.push_back(2);	// the count starts at 0, so index 2 is the 3rd component
	route.push_back(6);
	route.push_back(9);

	// Create cpd object
	cpd_Reroute rr( route );

	// Set up single parent to pass in
	nfg_Vecs paVecs(1);
	for ( int i = 0; i < 10; i++ )
		paVecs.at(0).push_back(i);

	//	Test sample_node()
	nfg_Vec outVec = rr.Sample_node( paVecs );
	Print_vec(outVec);

	//	Test Get_prob_prop()
	cout << "Prob: " << rr.Get_probability( outVec,  paVecs ) << endl;

	cout << "Unit test completed." << endl << endl;
}

