
// *************************************************************************
//	LibNFG: Net-form game library
//
//	Author:		Ritchie Lee
//
//  Copyright 2011 United States Government as represented by the Administrator 
//  of the National Aeronautics and Space Administration.  All Rights Reserved.  
//  Licensed under the Apache License, Version 2.0 (the "License"); you may not 
//  use this file except in compliance with the License.  You may obtain a copy 
//  of the License at http://www.apache.org/licenses/LICENSE-2.0
//  Unless required by applicable law or agreed to in writing, software 
//  distributed under the License is distributed on an "AS IS" BASIS, WITHOUT 
//  WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the 
//  License for the specific language governing permissions and limitations 
//  under the License.
//
//	Revision Log
//
//	Username	Date		Description
//	-------------------------------------------------------
//	rlee		20101025	Created
//	rlee		20110512	Added to Unit_test, and comments
//
// *************************************************************************


#include "cpd_tablebasic.h"

using namespace nfg;
using namespace boost;
using namespace boost::math;

// *************************************************************************
// *************************************************************************
cpd_TableBasic::cpd_TableBasic( const nfg_Dim& dims, const nfg_Vec& vals )
{
	//	Read-in order is [self pa0 pa1 ... ]
	//	p's should be incremental order as [0 0 0], [0 0 1], [0 1 0] ...
	//	See unit test for example usage

	// Check size implied by dims vs size of vals
	int totalSize = 1;
	foreach( unsigned int size, dims )
		totalSize *= size;

	// make sure the two match
	bassert( totalSize == vals.size() );

	//	Store the table in an NDArray
	NDArray tbl;
	tbl.Set_size ( dims );
	tbl.Set_data ( vals );

	//	Save it to member variables
	m_table		= tbl;

}
// *************************************************************************
// *************************************************************************
cpd_TableBasic::~cpd_TableBasic ( void )
{
}
// *************************************************************************
// *************************************************************************
nfg_Vec	cpd_TableBasic::Sample_node ( const nfg_Vecs& paVecs )
{	
	//	assert size of input vectors is correct
	bassert ( paVecs.size() == m_table.Get_num_dims()-1 );

	nfg_Vecs paVecs2( paVecs );	// make a copy
	
	// Treat dno as last value on node
	for( unsigned int i = 0; i < paVecs2.size(); i++ )
	{
		if( paVecs2.at(i).empty() )
		{
			paVecs2.at(i).push_back( m_table.Get_size(i+1) );
		}
	}

	//	assert each input size match
	for ( unsigned int i = 0; i < paVecs2.size(); i++ )
		bassert ( paVecs2.at(i).size() == 1 );

	nfg_Vec cum_distr;	// cumulative mass distribution
	double cmf_temp = 0.0;

	//	Use the parent values as the index to look up the 
	//	probabilities in the store probability table
	nfg_Dim idv;
	idv.push_back(0);
	for ( unsigned int i = 0; i < paVecs2.size(); i++ )
		idv.push_back( iround(paVecs2.at(i).at(0)) );

	//	Look up and accumulate a cumulative mass function for
	//	each possible output value
	for ( unsigned int i = 0; i < m_table.Get_size(0); i++ )
	{
		idv.at(0) = i;
		cmf_temp += m_table.Get_value( idv );
		cum_distr.push_back( cmf_temp );
	}

	//	Do a weighted sample based on the cumulative mass function
	//	and return the value
	return nfg_Vec(1,Weighted_sample ( cum_distr ));
}
// *************************************************************************
// *************************************************************************
double cpd_TableBasic::Get_probability ( const nfg_Vec& outVec, const nfg_Vecs& paVecs )
{	
	nfg_Vecs paVecs2( paVecs );	// make a copy
	
	// Treat dno as last value on node
	for( unsigned int i = 0; i < paVecs2.size(); i++ )
	{
		if( paVecs2.at(i).empty() )
		{
			
			paVecs2.at(i).push_back( m_table.Get_size(i+1)-1 );
		}
	}
	
	nfg_Dim idv;
	idv.push_back( iround(outVec.at(0)) );	

	//	Use the parent values as lookup index into the stored probability table
	for ( unsigned int i = 0; i < paVecs2.size(); i++ )
	{
		idv.push_back( iround(paVecs2.at(i).at(0)) );
	}

	//	return that probability value
	return m_table.Get_value( idv );
}

// *************************************************************************
// *************************************************************************
void cpd_TableBasic::Unit_test ( void )
{
	//	Set up cpd
	//	self has 3 possible values, pa0 has 2 (0 and dno), pa1 has 2 (0 and 1)
	int			T_dims[]	= { 3, 2, 2 };

	//	Rows: self
	//	P(self=0 | pa0=0,pa1=0)=0.3; P(self=0 | pa0=0,pa1=1)=0.9; 
	//	P(self=0 | pa0=dno,pa1=0)=0.05; P(self=0 | pa0=dno,pa1=1)=0.5;
	//	P(self=1 | pa0=0,pa1=0)=0.4; P(self=1 | pa0=0,pa1=1)=0.08; 
	//	P(self=1 | pa0=dno,pa1=0)=0.25; P(self=1 | pa0=dno,pa1=1)=0.3;
	//	P(self=2 | pa0=0,pa1=0)=0.3; P(self=2 | pa0=0,pa1=1)=0.02; 
	//	P(self=2 | pa0=dno,pa1=0)=0.7; P(self=2 | pa0=dno,pa1=1)=0.2;
	double		T_probs[]	= { 0.3, 0.9, 0.05, 0.5, 
								0.4, 0.08, 0.25, 0.3,
								0.3, 0.02, 0.7, 0.2 };

	// Construct cpd object
	cpd_TableBasic tab ( nfg_Dim(T_dims,T_dims+sizeof(T_dims)/sizeof(int)), 
						nfg_Vec(T_probs,T_probs+sizeof(T_probs)/sizeof(double)) );

	//	Load parents
	nfg_Vecs paVecs;
	paVecs.push_back( nfg_Vec(1,0.0) );	// pa0
	paVecs.push_back( nfg_Vec(1,1.0) );	// pa1

	//	Test Sample_node()
	int N = 2000;
	nfg_Dim hist(3,0);
	nfg_Vec outVec;
	// Sample N times and collect satistics.  Make sure the fraction matches the
	// respective probabilities in the CPD table
	for ( int i = 0; i < N; i++ )
	{
		outVec = tab.Sample_node ( paVecs );
		++hist.at( (int)outVec.at(0) );
	}
	cout << "Statistics:" << endl;
	cout << "0:" << (double)hist.at(0) / N << endl;
	cout << "1:" << (double)hist.at(1) / N << endl;
	cout << "2:" << (double)hist.at(2) / N << endl << endl;

	//	Test Get_probability()
	// Manually (and arbitrarily) choose some outVec and paVecs to test
	outVec.at(0) = 1; paVecs.at(0) = nfg_Vec(1,0); paVecs.at(1) = nfg_Vec(1,0);
	cout << "P(1|0,0): " << tab.Get_probability(outVec,paVecs) << endl; // P=0.4
	outVec.at(0) = 2; paVecs.at(0).clear(); paVecs.at(1) = nfg_Vec(1,0);
	cout << "P(2|1,0): " << tab.Get_probability(outVec,paVecs) << endl; // P=0.7
	outVec.at(0) = 0; paVecs.at(0) = nfg_Vec(1,0); paVecs.at(1) = nfg_Vec(1,1);
	cout << "P(0|0,1): " << tab.Get_probability(outVec,paVecs) << endl; // P=0.9
	outVec.at(0) = 1; paVecs.at(0).clear(); paVecs.at(1) = nfg_Vec(1,1);
	cout << "P(1|1,1): " << tab.Get_probability(outVec,paVecs) << endl;	// P=0.3

	cout << "Unit test completed." << endl << endl;
}

