
// *************************************************************************
//	LibNFG: Net-form game library
//
//	Author:		Ritchie Lee
//
//  Copyright 2011 United States Government as represented by the Administrator 
//  of the National Aeronautics and Space Administration.  All Rights Reserved.  
//  Licensed under the Apache License, Version 2.0 (the "License"); you may not 
//  use this file except in compliance with the License.  You may obtain a copy 
//  of the License at http://www.apache.org/licenses/LICENSE-2.0
//  Unless required by applicable law or agreed to in writing, software 
//  distributed under the License is distributed on an "AS IS" BASIS, WITHOUT 
//  WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the 
//  License for the specific language governing permissions and limitations 
//  under the License.
//
//	Revision Log
//
//	Username	Date		Description
//	-------------------------------------------------------
//	rlee		20110516	Created
//	rlee		20110816	Converted to inherited class
//
// *************************************************************************

#ifndef INCLUDED_CPD_TABLEVECTORFULL_H
#define INCLUDED_CPD_TABLEVECTORFULL_H

#include <boost/math/special_functions/round.hpp>
#include "cpd_tablebasic.h"
#include "ndarray.h"
#include "types.h"
#include "helper.h"
#include "cpd.h"

/**
 * cpd_TableVectorFull: CPD that supports vector discrete values.
 *		Supports full coupling of the components.
 *		For decoupled components, use cpd_TableVectorIndep.
 *		Works by re-indexing vectors down to a 1D problem that is
 *		passed to cpd_TableBasic.
 */
class cpd_TableVectorFull : public cpd
{
public:

	/**	Constructor */
	/**
	 * cpd_TableVectorFull: Constructor
	 *	space - a vector of vector of sizes. space = [[out] [pa0] [pa1] ...]
	 *		[out] = [[# possible values in out dim0] [# possible values in out dim1] ...]
	 *		[pa0] = [[# possible values in pa0 dim0] [# possible values in pa0 dim1] ...]
	 *	paHasDno - a vector of bools that indicate whether a parent has a dno value
	 *		[paHasDno] = [[pa0 hasDno] [pa1 hasDno] ...]
	 *	dimProbs - specifies the values of the CPD table for this output component
	 *		see unit_test for organization.
	 */
	cpd_TableVectorFull( const nfg_Dims& space, const nfg_Mask& paHasDno, 
													const nfg_Vec& dimProbs );

	/**	Destructor */
	virtual ~cpd_TableVectorFull();

	/// Variables

	///	Methods
	/**
	 * Sample_node: Sample the node given parents
	 *	paVecs - parent vectors
	 *	returns: sample value of node
	 */
	nfg_Vec	Sample_node ( const nfg_Vecs& paVecs );
	/**
	 * Get_probability: Get probability of a vector given parents
	 *	outVec - a sample/realization of the node
	 *	paVecs - parent vectors
	 *	returns: probability that this occurred
	 */
	double	Get_probability ( const nfg_Vec& outVec, const nfg_Vecs& paVecs );

	// Debug
	static void Unit_test ( void );

private:

	/// Variables
	nfg_Dims					m_space;		// Sizes of self and all the parents in each dimension
	nfg_Dim						m_elements;		// Number of elements in each of self, pa0, pa1, etc.
	nfg_Dims					m_strides;		// Stride length is used as the conversion factor between
												// N-d and 1-d indexing.
	nfg_Dim						m_ndims;		// Size of vec in each of self, pa0, pa1, etc.
	cpd_TableBasic*				mp_table;		// CPD table specifying all probabilities
	nfg_Mask					m_paHasDno;		// Vector [pa0 pa1...] specifying whether a parent has a dno value

	/// Methods	
	/**
	 * Check_bounds: Checks to make sure that a given index is valid. i.e., within the
	 *		size bounds of what is declared.
	 *	dim - dimension of interest. 0 is self, 1 is pa0, 2 is pa1, etc.
	 *	idv - index to check
	 *	returns: true if within bounds/valid, false otherwise
	 */
	bool Check_bounds( const unsigned int dim, const nfg_Dim& idv );
	/**
	 * Get_1D_index: Convert an N-d index to a 1-d index for a center dimension
	 *	dim - dimension of interest. 0 is self, 1 is pa0, 2 is pa1, etc.
	 *	idv - N-d index to convert
	 *	return: 1-d index
	 */
	unsigned int Get_1d_index ( const unsigned int dim, const nfg_Dim& idv );
	/**
	 * Convert_vecs_to_basic: Convert from vectored representation to 1-d representation
	 *		for interacting with cpd_TableBasic
	 *	vecs - a vector of vecs (such as paVecs) to be converted
	 *	returns: vector of converted "basic" representation. i.e., data is in .at(0) only.
	 */
	nfg_Vecs	Convert_vecs_to_basic ( const nfg_Vecs& vecs );
	/**
	 * Convert_basic_to_vec: Convert from 1-d "basic" representation to N-d representation
	 *		suitable for output
	 *	basic - vector in basic representation. i.e., data is in .at(0) only.
	 *	returns: vector in N-d representation
	 */
	nfg_Vec Convert_basic_to_vec ( const nfg_Vec& basic );
};

#endif

