
// *************************************************************************
//	LibNFG: Net-form game library
//
//	Author:		Ritchie Lee
//
//  Copyright 2011 United States Government as represented by the Administrator 
//  of the National Aeronautics and Space Administration.  All Rights Reserved.  
//  Licensed under the Apache License, Version 2.0 (the "License"); you may not 
//  use this file except in compliance with the License.  You may obtain a copy 
//  of the License at http://www.apache.org/licenses/LICENSE-2.0
//  Unless required by applicable law or agreed to in writing, software 
//  distributed under the License is distributed on an "AS IS" BASIS, WITHOUT 
//  WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the 
//  License for the specific language governing permissions and limitations 
//  under the License.
//
//	Revision Log
//
//	Username	Date		Description
//	-------------------------------------------------------
//	rlee		20110401	Created
//
// *************************************************************************

#include "if_bayesnettoolbox.h"

#include <string>
#include <iostream>
#include <fstream>

using namespace std;
using namespace boost;

// *************************************************************************
// *************************************************************************
if_BayesNetToolbox::if_BayesNetToolbox( void )
{
}
// *************************************************************************
// *************************************************************************
if_BayesNetToolbox::~if_BayesNetToolbox ( void )
{
}
// *************************************************************************
// *************************************************************************
void if_BayesNetToolbox::Draw_graph( const string filename, BayesNet& bnet, 
												bool append, const string dagName,
												const string labelsName )
{
	// Open a file stream
	ofstream mfile;

	// If append, use the append flag and immediately insert an endl.
	if ( append )
	{
		mfile.open( filename.c_str(), ios_base::app );
		mfile << endl;
	}
	// If overwrite, open the file
	else
		mfile.open( filename.c_str() );

	if ( mfile.is_open() )
	{
		// For convenience,
		int N	= bnet.Get_num_vertices();

		// Create a matrix of size N by N of zeros
		mfile << dagName << "=zeros(" << N << "," << N << ");" << endl;
		
		// For each edge in the graph, make mark the corresponding entry in the matrix
		// with a 1. 1 means connected, 0 means not connected.
		nfg_EdgeIter ei, ei_end;
		for ( tie(ei,ei_end) = edges(bnet.Get_graph()); ei != ei_end; ++ei )
		{
			mfile << dagName << "(" << source(*ei, bnet.Get_graph())+1
					<< "," << target(*ei,bnet.Get_graph())+1 << ")=1;" << endl;
		}
		// Output the vertex labels into a cell array
		foreach ( nfg_Vertex vi, vertices(bnet.Get_graph()) )
		{
			mfile << labelsName << "{" << vi+1 << "}='" << bnet.Get_vertex_name(vi) 
							<< "';" << endl;
		}
		// Call draw_graph which is a part of BNT
		mfile << "draw_graph(" << dagName << "," << labelsName << ");" << endl;

		///	Deal with special characters in the string
		string bname = Replace_Matlab_special_chars( bnet.Get_bnet_name() );
		
		// Use the bnet name as the title of the figure
		mfile << "title('" << bname << "');" << endl;

		// Close the file
		mfile.close();
	}
	else
	{
		cout << "\n\nERROR! Could not open output file: " << filename << endl;
		bassert( false );
	}

	return;
}
// *************************************************************************
// *************************************************************************
void if_BayesNetToolbox::Draw_graph( const string filename, IBayesNet& ibnet, 
												bool append, const string dagName,
												const string labelsName )
{
	// Open a file stream
	ofstream mfile;

	// If append, use the append flag and immediately insert an endl.
	if ( append )
	{
		mfile.open( filename.c_str(), ios_base::app );
		mfile << endl;
	}
	// If overwrite, open the file
	else
		mfile.open( filename.c_str() );

	if ( mfile.is_open() )
	{
		// For convenience,
		int N_base	= ibnet.Get_base_bnet().Get_num_vertices();
		int N_kernel = ibnet.Get_kernel_bnet().Get_num_vertices();
		int N = N_base + 2*N_kernel;

		// Create a matrix of size N by N of zeros
		mfile << dagName << "=zeros(" << N << "," << N << ");" << endl;
		
		// For each edge in the base bnet, mark the corresponding entry in the matrix
		// with a 1. 1 means connected, 0 means not connected.
		nfg_EdgeIter ei1, ei1_end;
		for ( tie(ei1,ei1_end) = edges(ibnet.Get_base_bnet().Get_graph()); ei1 != ei1_end; ++ei1 )
		{
			mfile << dagName << "(" << source(*ei1, ibnet.Get_base_bnet().Get_graph())+1
					<< "," << target(*ei1,ibnet.Get_base_bnet().Get_graph())+1 << ")=1;" << endl;
		}
		// For each edge in the kernel bnet (first slice), mark the corresponding entry in the matrix
		// with a 1. 1 means connected, 0 means not connected.
		nfg_EdgeIter ei2, ei2_end;
		for ( tie(ei2,ei2_end) = edges(ibnet.Get_kernel_bnet().Get_graph()); ei2 != ei2_end; ++ei2 )
		{
			mfile << dagName << "(" << source(*ei2, ibnet.Get_kernel_bnet().Get_graph())+1+N_base
					<< "," << target(*ei2,ibnet.Get_kernel_bnet().Get_graph())+1+N_base << ")=1;" << endl;
		}
		// For each edge in the kernel bnet (second slice), mark the corresponding entry in the matrix
		// with a 1. 1 means connected, 0 means not connected.
		nfg_EdgeIter ei2b, ei2b_end;
		for ( tie(ei2b,ei2b_end) = edges(ibnet.Get_kernel_bnet().Get_graph()); ei2b != ei2b_end; ++ei2b )
		{
			mfile << dagName << "(" << source(*ei2b, ibnet.Get_kernel_bnet().Get_graph())+1+2*N_base
					<< "," << target(*ei2b,ibnet.Get_kernel_bnet().Get_graph())+1+2*N_base << ")=1;" << endl;
		}
		// For each interconnecting edge, mark the corresponding entry in the matrix
		// with a 1. 1 means connected, 0 means not connected.
		foreach ( nfg_Edge ei3, ibnet.Get_base2kernelEdges() )
		{
			mfile << dagName << "(" << ei3.first+1
					<< "," << ei3.second+1+N_base << ")=1;" << endl;
		}
		// For each interconnecting edge, mark the corresponding entry in the matrix
		// with a 1. 1 means connected, 0 means not connected.
		foreach ( nfg_Edge ei3b, ibnet.Get_base2kernelEdges() )
		{
			mfile << dagName << "(" << ei3b.first+1+N_base
					<< "," << ei3b.second+1+2*N_base << ")=1;" << endl;
		}
		// Output the base vertex labels into a cell array
		foreach ( nfg_Vertex vi, vertices(ibnet.Get_base_bnet().Get_graph()) )
		{
			mfile << labelsName << "{" << vi+1 << "}='" << ibnet.Get_base_bnet().Get_vertex_name(vi) 
				<<"(0)"	<< "';" << endl;
		}
		// Output the kernel vertex labels (first slice) into a cell array
		foreach ( nfg_Vertex vi, vertices(ibnet.Get_kernel_bnet().Get_graph()) )
		{
			mfile << labelsName << "{" << vi+1+N_base << "}='" << ibnet.Get_kernel_bnet().Get_vertex_name(vi) 
					<<"(1)"	<< "';" << endl;
		}
		// Output the kernel vertex labels (second slice) into a cell array
		foreach ( nfg_Vertex vi, vertices(ibnet.Get_kernel_bnet().Get_graph()) )
		{
			mfile << labelsName << "{" << vi+1+2*N_base << "}='" << ibnet.Get_kernel_bnet().Get_vertex_name(vi) 
					<<"(2)"	<< "';" << endl;
		}
		// Call draw_graph which is a part of BNT
		mfile << "draw_graph(" << dagName << "," << labelsName << ");" << endl;

		///	Deal with special characters in the string
		string bname = Replace_Matlab_special_chars( ibnet.Get_ibnet_name() );
		
		// Use the bnet name as the title of the figure
		mfile << "title('" << bname << "');" << endl;

		// Close the file
		mfile.close();
	}
	else
	{
		cout << "\n\nERROR! Could not open output file: " << filename << endl;
		bassert( false );
	}

	return;
}



