
// *************************************************************************
//	LibNFG: Net-form game library
//
//	Author:		Ritchie Lee
//
//  Copyright 2011 United States Government as represented by the Administrator 
//  of the National Aeronautics and Space Administration.  All Rights Reserved.  
//  Licensed under the Apache License, Version 2.0 (the "License"); you may not 
//  use this file except in compliance with the License.  You may obtain a copy 
//  of the License at http://www.apache.org/licenses/LICENSE-2.0
//  Unless required by applicable law or agreed to in writing, software 
//  distributed under the License is distributed on an "AS IS" BASIS, WITHOUT 
//  WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the 
//  License for the specific language governing permissions and limitations 
//  under the License.
//
//	Revision Log
//
//	Username	Date		Description
//	-------------------------------------------------------
//	rlee		20110420	Created
//	rlee		20110602	Updated to current definition of branch nodes
//
// *************************************************************************

#ifndef INCLUDED_IF_GAMBIT_H
#define INCLUDED_IF_GAMBIT_H

#include <boost/optional.hpp>
#include "types.h"
#include "helper.h"
#include "bayesnet.h"

//	Interface to Matlab
class if_Gambit
{
public:

	/**	Constructor */
	if_Gambit( void );

	/**	Destructor */
	virtual ~ if_Gambit();

	/// Variables

	///	Methods
	static void		Export_to_efg_format ( BayesNet& bnet, const string filename );

	// Debug
	static void		Unit_subtest1( void );

private:

	/// Variables


	/// Methods
	/**
	 * Expand_node: Recursive function that will call (and expand) itself on each of its
	 *		possible states.  Function body writes each body/node line of efg file
	 *	outfile - file stream to output to
	 *	bnet - nfg object
	 *	v - current vertex id
	 *	prevInst - partial instantiation tracking values sampled so far
	 *	tNum - central variable to give unique terminal numbers
	 *	cNum - central variable to give unique chance infoset numbers
	 *	dnodes - list of decision nodes
	 */
	static void		Expand_node ( ofstream& outfile, BayesNet& bnet, const nfg_Vertex v, 
							const nfg_OptInst& prevInst, unsigned int& tNum,
							unsigned int& cNum,
							const nfg_Vertices& dNodes );
	/**
	 * Find_value: Get index of the first element in list with value matching val.
	 *	vlist - list to be searched
	 *	val - value to be found in the list
	 *	returns: an optional index.  uninitialized for not found. otherwise contains index.
	 */
	static optional<int> 
					Find_value( const nfg_Vertices& vlist, const nfg_Vertex val );
	/**
	 * Write_line_chance: Write a chance formatted line to an efg file
	 *	outfile - file stream to output to
	 *	bnet - nfg object
	 *	v - current vertex id
	 *	prevInst - partial instantiation tracking values sampled so far
	 *	cNum - central variable to give unique chance infoset numbers
	 *	dnodes - list of decision nodes
	 */
	static void Write_line_chance( ofstream& outfile, BayesNet& bnet, const nfg_Vertex v,
								  const nfg_OptInst& prevInst, unsigned int& cNum,
								  const nfg_Vertices& dNodes );
	/**
	 * Write_line_player: Write a player formatted line to an efg file
	 *	outfile - file stream to output to
	 *	bnet - nfg object
	 *	v - current vertex id
	 *	prevInst - partial instantiation tracking values sampled so far
	 *	pNum - player infoset number to be written
	 *	dnodes - list of decision nodes
	 */
	static void Write_line_player( ofstream& outfile, BayesNet& bnet, const nfg_Vertex v,
								  const nfg_OptInst& prevInst,
								  unsigned int& pNum, const nfg_Vertices& dNodes );
	/**
	 * Write_line_terminal: Write a terminal formatted line to an efg file
	 *	outfile - file stream to output to
	 *	bnet - nfg object
	 *	v - current vertex id
	 *	prevInst - partial instantiation tracking values sampled so far
	 *	tNum - central variable to give unique terminal numbers
	 *	dnodes - list of decision nodes
	 */
	static void Write_line_terminal( ofstream& outfile, BayesNet& bnet, const nfg_Vertex v,
								  const nfg_OptInst& prevInst, unsigned int& tNum,
								  const nfg_Vertices& dNodes );
	/**
	 * Get_parent_values: Get values of parents for a node v
	 *	bnet - nfg object
	 *	inst - could be an instantiation or partial instantiation containing the values of the parents of v
	 *	v - vertex id
	 *	returns: a vector (size num parents of v) of vectors containing the values of the parents
	 *		ordered by ascending parent vertex id.
	 */
	static nfg_Vecs		Get_parent_values( BayesNet& bnet, const nfg_OptInst& optInst, 
										const nfg_Vertex& v );

private:
	/**
	 * DiscreteSpaceIterator: Iterates integers over a multi-dimensional discrete space
	 */
	class DiscreteSpaceIterator
	{

	public:
		/**
		 * DiscreteSpaceIterator: Constructor
		 *	space - discrete space definition according to nfg_Space format
		 */
		DiscreteSpaceIterator ( const nfg_Space& space );
		virtual ~DiscreteSpaceIterator(){};
		
		/**
		 * operator++: iterator increment
		 */
		void operator++( void );
		/**
		 * operator(): retrieves the object currently pointed to
		 *	returns: nfg_Vec object currently pointed to
		 */
		nfg_Vec operator()( void );
	private:
		nfg_Dim		m_index;	// points to the current place in the iteration
		nfg_Dim		m_end;		// points to the end in each dimesion so you know when to
								// loop back
	
	};

};

#endif

