
// *************************************************************************
//	LibNFG: Net-form game library
//
//	Author:		Ritchie Lee
//
//  Copyright 2011 United States Government as represented by the Administrator 
//  of the National Aeronautics and Space Administration.  All Rights Reserved.  
//  Licensed under the Apache License, Version 2.0 (the "License"); you may not 
//  use this file except in compliance with the License.  You may obtain a copy 
//  of the License at http://www.apache.org/licenses/LICENSE-2.0
//  Unless required by applicable law or agreed to in writing, software 
//  distributed under the License is distributed on an "AS IS" BASIS, WITHOUT 
//  WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the 
//  License for the specific language governing permissions and limitations 
//  under the License.
//
//	Revision Log
//
//	Username	Date		Description
//	-------------------------------------------------------
//	rlee		20110523	Created
//
// *************************************************************************

#include "if_mpi.h"

#include <boost/function.hpp>
#include <boost/lambda/bind.hpp>
#include <iostream>
#include <time.h>
#include <boost/serialization/vector.hpp>
#include <boost/serialization/list.hpp>

#ifdef MPI_ENABLE
#include <boost/mpi.hpp>
namespace mpi = boost::mpi;
#endif

using namespace std;
using namespace nfg;
using namespace boost::lambda;


// *************************************************************************
// *************************************************************************
if_Mpi::if_Mpi( void )
{
}
// *************************************************************************
// *************************************************************************
if_Mpi::~if_Mpi ( void )
{
}
// Pre-processor that enables the MPI support
#ifdef MPI_ENABLE
// *************************************************************************
// *************************************************************************
void if_Mpi::Parallel_exec( const function<nfg_WtdInsts (void)>& fo_parFunc,
							const function<void (vector<nfg_WtdInsts>&)>& fo_postFunc )
{
	// Declare mpi system variables
	mpi::environment env;
	mpi::communicator world;	// responsible for all communication between nodes

	// need to make the seed dependent on world rank, 
	// otherwise results may be repeated between
	// compute nodes due to having the same random seed
	srand((unsigned int)time(0) + world.rank());

	// Declare a vector to contain all the gathered info
	vector<nfg_WtdInsts> all;

	// Call parallel processing user-specified function
	nfg_WtdInsts each = fo_parFunc();

	// Head node executes this...
	if ( world.rank() == 0 )
	{
		// Collect all results into a vector at the head node
		mpi::gather( world, each, all, 0 );
		
		// Execute user-specified post processing function at the head node
		// giving it the output from all compute node
		fo_postFunc( all );
	}
	// Other nodes execute this...
	else
	{
		// Send it back to head node
		mpi::gather( world, each, 0 );
	}

	
	return;
}
// Disable MPI support
#else
// *************************************************************************
// *************************************************************************
void if_Mpi::Parallel_exec( const function<nfg_WtdInsts (void)>& fo_parFunc,
							const function<void (vector<nfg_WtdInsts>&)>& fo_postFunc )
{

	// Declare a vector to contain all the gathered info
	vector<nfg_WtdInsts> all;

	// Call parallel processing user-specified function
	all.push_back( fo_parFunc() );

	// Execute user-specified post processing function at the head node
	// giving it the output from all compute node
	fo_postFunc( all );
	
	return;
}
#endif
// *************************************************************************
// *************************************************************************
void if_Mpi::Unit_test( void )
{
	cout << "Starting if_Mpi unit test..." << endl;
	
	// Test parallel execution
	if_Mpi::Parallel_exec( bind(&Unit_test_parFunc), bind(&Unit_test_postFunc,_1) );

	cout << "Unit test completed!" << endl << endl;

}
// *************************************************************************
// *************************************************************************
nfg_WtdInsts if_Mpi::Unit_test_parFunc( void )
{
	// Parallel process function (one instance on each compute node)
	// Distribute computing
	cout << "Parallel processing..." 
			<< endl;

	nfg_WtdInsts wInsts;
	
	// Collect 5 samples from each compute node
	int N = 5;
	for( int i = 0; i < N; i++ )
	{
		// Some unique test object.  Details not important.
		nfg_WtdInst wInst;
		wInst.m_weight	= Rand_uniform_01();
		wInst.m_inst	= nfg_Inst();
		wInsts.push_back( wInst );
	}

	return wInsts;
}
// *************************************************************************
// *************************************************************************
void if_Mpi::Unit_test_postFunc( const vector<nfg_WtdInsts>& results )
{
	// Post process function (only on head compute node)
	// For compiling data at the end
	cout << "Post processing..." << endl;
	
	nfg_WtdInsts finalList;
	
	// Re-packages the sample list into standard format for outputting
	foreach( const nfg_WtdInsts& wInsts, results )
	{
		foreach( const nfg_WtdInst& wInst, wInsts )
		{
			finalList.push_back( wInst );
		}
	}

	return;
}

