
// *************************************************************************
//	LibNFG: Net-form game library
//
//	Author:		Ritchie Lee
//
//  Copyright 2011 United States Government as represented by the Administrator 
//  of the National Aeronautics and Space Administration.  All Rights Reserved.  
//  Licensed under the Apache License, Version 2.0 (the "License"); you may not 
//  use this file except in compliance with the License.  You may obtain a copy 
//  of the License at http://www.apache.org/licenses/LICENSE-2.0
//  Unless required by applicable law or agreed to in writing, software 
//  distributed under the License is distributed on an "AS IS" BASIS, WITHOUT 
//  WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the 
//  License for the specific language governing permissions and limitations 
//  under the License.
//
//	Revision Log
//
//	Username	Date		Description
//	-------------------------------------------------------
//	rlee		20101025	Created
//
// *************************************************************************

#include "ndarray.h"

// *************************************************************************
// *************************************************************************
NDArray::NDArray( void )
{	
}
// *************************************************************************
// *************************************************************************
NDArray::~NDArray ( void )
{
}
// *************************************************************************
// *************************************************************************
void NDArray::Set_size( const vector<unsigned int>& sizes )
{
	m_sizes		= sizes;			// save a copy
	m_ndims		= sizes.size();		// number of dimensions
	
	// number of elements is the product of sizes in each dimensions
	m_elements = 1;
	for ( unsigned int i = 0; i < m_ndims; i++ )
		m_elements *= sizes.at(i);

	// Figure out the stride length in each dimension
	for ( unsigned int i = 0; i < m_ndims; i++ )
	{
		unsigned int stride = 1;
		for ( unsigned int j = i+1; j < m_ndims; j++ )
		{
			stride *= sizes.at(j);
		}
		m_strides.push_back(stride);
	}
	
	// resize one dimensional array to contain all the elements
	// and initialize all values to zero
	m_data.resize( m_elements, 0 );
}
// *************************************************************************
// *************************************************************************
double NDArray::Get_value( const vector<unsigned int>& idv )
{
	// Convert to linear indexing and look up the value
	return m_data.at(Get_index(idv));	
}
// *************************************************************************
// *************************************************************************
void NDArray::Set_value( const vector<unsigned int>& idv, const double val )
{
	// Convert to linear indexing and set the value
	Set_value( Get_index(idv),val );
}
// *************************************************************************
// *************************************************************************
unsigned int NDArray::Get_index( const vector<unsigned int>& idv )
{
	// First, make sure the request is valid
	assert ( Check_bounds(idv) );

	// Convert tuple to linear indexing by multiplying by stride lengths
	unsigned int idx = 0;
	for ( unsigned int i = 0; i < m_ndims; i++ )
	{
		idx += idv.at(i)*m_strides.at(i);
	}

	return idx;
}
// *************************************************************************
// *************************************************************************
bool NDArray::Check_bounds( const vector<unsigned int>& idv )
{
	bool isValid = true;

	// In order to be within bounds, it needs to be between 0 and size
	// in each dimension
	for ( unsigned int i = 0; i < m_ndims; i++ )
		isValid = isValid && ( idv[i] < m_sizes[i] ) && ( idv[i] >= 0 );

	return isValid;
}
// *************************************************************************
// *************************************************************************
void NDArray::Set_value ( const unsigned int idx, const double val )
{
	// Set a value by linear index
	m_data.at(idx) = val;
}
// *************************************************************************
// *************************************************************************
void NDArray::Set_data ( const vector<double>& data )
{
	m_data = data;
}
