
// *************************************************************************
//	LibNFG: Net-form game library
//
//	Author:		Ritchie Lee
//
//  Copyright 2011 United States Government as represented by the Administrator 
//  of the National Aeronautics and Space Administration.  All Rights Reserved.  
//  Licensed under the Apache License, Version 2.0 (the "License"); you may not 
//  use this file except in compliance with the License.  You may obtain a copy 
//  of the License at http://www.apache.org/licenses/LICENSE-2.0
//  Unless required by applicable law or agreed to in writing, software 
//  distributed under the License is distributed on an "AS IS" BASIS, WITHOUT 
//  WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the 
//  License for the specific language governing permissions and limitations 
//  under the License.
//
//	Revision Log
//
//	Username	Date		Description
//	-------------------------------------------------------
//	rlee		20110407	Created
//
// *************************************************************************

#include "net_bobvsalice.h"

#include <boost/lambda/bind.hpp>
#include <boost/math/special_functions/round.hpp>
#include "bayesnet.h"
#include "alg_levelkrelaxed.h"
#include "cpd.h"
#include "cpd_tablebasic.h"
#include "cpd_decision.h"
#include "player.h"
#include "if_matlab.h"
#include "if_bayesnettoolbox.h"
#include "if_gambit.h"

using namespace nfg;
using namespace std;
using namespace boost;
using namespace boost::lambda;

// *************************************************************************
// *************************************************************************
void net_BobVsAlice::Run ( void )
{
	//*********************************************
	///	 Set up Bayes Net

	// Specify vertex names
	string vertexNames[] = 
	{	"C", "A1", "B1", "B2", "A2"	};

	// Define Edges of the graph
	nfg_Edge edgeArray[] = 
	{ 
		nfg_Edge(C,A1), nfg_Edge(C,B2), 
		nfg_Edge(A1,B1), nfg_Edge(B2,A2) 
	};

	// Specify vertex type
	nfg_eVertexType vertexTypes[] =
	{	CHANCE, DECISION, DECISION, DECISION, DECISION	};

	// Specify whether vertex is a branch node
	bool vertexIsBranch[] =
	{ true, false, false, false, false };

	// Specify which vertices have dno values
	bool vertexHasDno[] =
	{ false, true, true, true, true };

	// Specify the space of each vertex
	nfg_Spaces spaces(N);
	spaces.at(C).push_back( Construct_discrete_space1d ( Vectorize(0,1) ));
	spaces.at(A1).push_back( Construct_discrete_space1d ( Vectorize(0,1) ));
	spaces.at(B1).push_back( Construct_discrete_space1d ( Vectorize(0,1) ));
	spaces.at(A2).push_back( Construct_discrete_space1d ( Vectorize(0,1) ));
	spaces.at(B2).push_back( Construct_discrete_space1d ( Vectorize(0,1) ));

	////	Make Bayes Net
	// Pass all specified values above to the bnet constructor
	BayesNet bnet( "Bob vs Alice", N,
		nfg_Edges(edgeArray,edgeArray+sizeof(edgeArray)/sizeof(nfg_Edge)), 
		nfg_Dim(N,1),			// all sizes are 1
		spaces,
		nfg_Mask(N,false),		// all discrete nodes
		nfg_eVertexTypes(vertexTypes,vertexTypes+N),
		nfg_Mask(vertexIsBranch,vertexIsBranch+N),
		nfg_Mask(vertexHasDno, vertexHasDno+N) );

	// Set vertex names
	bnet.Set_vertex_names ( vector<string>(vertexNames,vertexNames+N) );

	// Set dno functions
	bnet.Set_dno_function( A1, bind(&Dno_function_AliceFirst,_1) );
	bnet.Set_dno_function( B1, bind(&Dno_function_AliceFirst,_1) );
	bnet.Set_dno_function( A2, bind(&Dno_function_BobFirst,_1) );
	bnet.Set_dno_function( B2, bind(&Dno_function_BobFirst,_1) );

	// Declare players
	Player Alice ( bind(&Util_function_Alice, _1), "Alice" );	// pass in utility function
	Player Bob ( bind(&Util_function_Bob, _1), "Bob" );		// pass in utility function
	// Specify which decision nodes are controlled by which players
	Alice.Add_controlled_node(A1);
	Alice.Add_controlled_node(A2);
	Bob.Add_controlled_node(B1);
	Bob.Add_controlled_node(B2);

	// Associate players with bnet
	bnet.Add_player(Alice);
	bnet.Add_player(Bob);

	// Create cpds
	// C has a tabular cpd
	//	Tails: C = 0; Heads: C = 1
	int			C_dims[]	= { 2 };			// 2 possible states
	double		C_probs[]	= { 0.5, 0.5 };		// 50-50 chance
	cpd_TableBasic		C_cpd ( nfg_Dim(C_dims,C_dims+sizeof(C_dims)/sizeof(int)), 
							nfg_Vec(C_probs,C_probs+sizeof(C_probs)/sizeof(double)));
	// Others are decision nodes.  Specify satisficing distributions.
	cpd_Decision	A1_cpd ( Alice );	// associate with player
	cpd_Decision	B1_cpd ( Bob );		// associate with player
	cpd_Decision	A2_cpd ( Alice );	// associate with player
	cpd_Decision	B2_cpd ( Bob );		// associate with player

	//	Associate cpds to bnet
	bnet.Set_cpd( C, &C_cpd );
	bnet.Set_cpd( A1, &A1_cpd );
	bnet.Set_cpd( A2, &A2_cpd );
	bnet.Set_cpd( B1, &B1_cpd );
	bnet.Set_cpd( B2, &B2_cpd );

	//	Error check
	bassert ( bnet.Check() );

	// Draw graph (to make sure the code understood your graph correctly)
	if_BayesNetToolbox::Draw_graph( "Output\\BvA_draw.m", bnet );

	//**************************************************
	///	Run main program
	
	// Choose custom analysis to run on bnet
	Analysis1(bnet);
	//Analysis2(bnet);

	return;
}
// *************************************************************************
// *************************************************************************
double net_BobVsAlice::Util_function_Alice ( const nfg_Inst& instVecs )
{

	//	Alice is trying to be different from Bob
	double u = 0;

	//	If heads
	if ( instVecs.at(C).at(0) == 0 )
	{
		if ( instVecs.at(A1).at(0) == instVecs.at(B1).at(0) )
			u = -1.0;
		else
			u = 1.0;
	}
	//	If tails
	else
	{
		if ( instVecs.at(A2).at(0) == instVecs.at(B2).at(0) )
			u = -1.0;
		else
			u = 1.0;
	}

	return u;
}
// *************************************************************************
// *************************************************************************
double net_BobVsAlice::Util_function_Bob ( const nfg_Inst& instVecs )
{
	//	Bob is trying to be same as Alice
	double u = 0;

	//	If tails
	if ( instVecs.at(C).at(0) == 0 )
	{
		if ( instVecs.at(A1).at(0) == instVecs.at(B1).at(0) )
			u = 1.0;
		else
			u = -1.0;
	}
	//	If heads
	else
	{
		if ( instVecs.at(A2).at(0) == instVecs.at(B2).at(0) )
			u = 1.0;
		else
			u = -1.0;
	}

	return u;
}
// *************************************************************************
// *************************************************************************
nfg_Vec net_BobVsAlice::Sample_movespace ( const nfg_Vecs& paVecs )
{
	//	The parent vectors are ignored

	// Satisficing distribution is simply random 0 or 1
	return nfg_Vec(1, Rand_int(0,1) );
}
// *************************************************************************
// *************************************************************************
nfg_Vertex net_BobVsAlice::Partition_function_C ( const nfg_Vec& vec )
{
	nfg_Vertex v;

	//	If tails, specify A1 as active branch
	if ( math::iround(vec.at(0)) == 0 )
	{
		v = A1;
	}
	//	If heads, specify B2 as active branch
	else
	{
		v = B2;
	}
	
	return v;
}
// *************************************************************************
// *************************************************************************
bool net_BobVsAlice::Dno_function_AliceFirst ( const nfg_OptInst& optInstVecs )
{
	// If heads, return dno
	if ( (*optInstVecs.at(C)).at(0) == 1 )
		return true;
	// If tails, return not dno
	else
		return false;
}
// *************************************************************************
// *************************************************************************
bool net_BobVsAlice::Dno_function_BobFirst ( const nfg_OptInst& optInstVecs )
{
	// If tails, return dno
	if ( (*optInstVecs.at(C)).at(0) == 0 )
		return true;
	// If heads, return not dno
	else
		return false;
}
// *************************************************************************
// *************************************************************************
nfg_Vec net_BobVsAlice::Sample_level0 ( const nfg_Vecs& paVecs )
{
	//	The parent vectors are ignored

	// level0 distribution is just random 0 or 1
	return nfg_Vec(1, Rand_int(0,1) );
}
// *************************************************************************
// *************************************************************************
void net_BobVsAlice::Analysis1 ( BayesNet& bnet )
{
	cout << "Running Analysis1..." << endl;

	// Declare inference engine
	alg_LevelKRelaxed engine;

	// Specify level0 distributions (functions) for each decision node
	nfg_SampleFuncs fo_Sample_level0_inst(N);
	fo_Sample_level0_inst.at(A1) = (nfg_SampleFunc)bind(&Sample_level0, _1);
	fo_Sample_level0_inst.at(A2) = (nfg_SampleFunc)bind(&Sample_level0, _1);
	fo_Sample_level0_inst.at(B1) = (nfg_SampleFunc)bind(&Sample_level0, _1);
	fo_Sample_level0_inst.at(B2) = (nfg_SampleFunc)bind(&Sample_level0, _1);

	// Specify satisficing distributions (functions) for each decision node
	nfg_SampleFuncs fo_Sample_satisf_inst(N);
	fo_Sample_satisf_inst.at(A1) = bind(&Sample_movespace,_1);	// specify satisficing distr.
	fo_Sample_satisf_inst.at(A2) = bind(&Sample_movespace,_1);	// specify satisficing distr.
	fo_Sample_satisf_inst.at(B1) = bind(&Sample_movespace,_1);	// specify satisficing distr.
	fo_Sample_satisf_inst.at(B2) = bind(&Sample_movespace,_1);	// specify satisficing distr.

	// Set level K parameters
	engine.Set_params( bnet, alg_LevelKRelaxed::D_RELAXED, 5, 10, fo_Sample_level0_inst,
								fo_Sample_satisf_inst,
								alg_LevelKRelaxed::BEST, true );
	
	// Used to contain the samples
	nfg_WtdInsts wInsts;

	// Collect 500 samples of the net
	for ( int i = 0; i < 500; i++ )
	{
		cout << i << endl;		// Progress indicator to console
		// Generate a sample and add it to the list
		wInsts.push_back( engine.LW_forward_sample( bnet, 0 ) );
	}
	// Output the samples to a Matlab file for post-processing
	if_Matlab::Export_WtdInsts_to_file ( "Output\\BvA_vars.m", "samples", wInsts, false );

	//	Track players' utility distribution
	nfg_Vec AliceUtils, BobUtils;
	foreach ( nfg_WtdInst wInst, wInsts )
	{
		// Map each instantiation (sample) of the net to the player's utilities
		// and save them to a list
		AliceUtils.push_back( Util_function_Alice(wInst.m_inst) );
		BobUtils.push_back( Util_function_Bob(wInst.m_inst) );
	}
	// Output the utility values to a Matlab file for post-processing
	if_Matlab::Export_Vec_to_file ( "Output\\BvA_vars.m", "AliceUtils", AliceUtils, true );
	if_Matlab::Export_Vec_to_file ( "Output\\BvA_vars.m", "BobUtils", BobUtils, true );
}
// *************************************************************************
// *************************************************************************
void net_BobVsAlice::Analysis2 ( BayesNet& bnet )
{
	// Export to GAMBIT to perform QRE analysis
	cout << "Running Analysis2..." << endl;
	if_Gambit::Export_to_efg_format( bnet, "Output\\BvA.efg" );
}


