
// *************************************************************************
//	LibNFG: Net-form game library
//
//	Author:		Ritchie Lee
//
//  Copyright 2011 United States Government as represented by the Administrator 
//  of the National Aeronautics and Space Administration.  All Rights Reserved.  
//  Licensed under the Apache License, Version 2.0 (the "License"); you may not 
//  use this file except in compliance with the License.  You may obtain a copy 
//  of the License at http://www.apache.org/licenses/LICENSE-2.0
//  Unless required by applicable law or agreed to in writing, software 
//  distributed under the License is distributed on an "AS IS" BASIS, WITHOUT 
//  WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the 
//  License for the specific language governing permissions and limitations 
//  under the License.
//
//	Revision Log
//
//	Username	Date		Description
//	-------------------------------------------------------
//	rlee		20110407	Created
//
// *************************************************************************

#ifndef INCLUDED_NET_BOBVSALICE_H
#define INCLUDED_NET_BOBVSALICE_H

#include "helper.h"
#include "types.h"
#include "bayesnet.h"

/**
 * "Bob vs Alice": Example 3 from Learning in Net Form Games (Wolpert)
 *	Chance node selects which player goes first (branching) in a game.
 *	Demonstrates branching
 */
class net_BobVsAlice
{
	// Make convenient labels for the vertices
	enum { C, A1, B1, B2, A2, N };

public:
	/**
	 * Run: Execute example sequence
	 */
	static void			Run ( void );
	/**
	 * Util_function_Alice: Alice's utility function
	 *	instVecs: an instantiation of the net
	 *	returns: utility value
	 */
	static double		Util_function_Alice ( const nfg_Inst& instVecs );
	/**
	 * Util_function_Bob: Bob's utility function
	 *	instVecs: an instantiation of the net
	 *	returns: utility value
	 */
	static double		Util_function_Bob ( const nfg_Inst& instVecs );
	/**
	 * Sample_movespace: Sample a possible move of a decision node 
	 *	(satisficing distribution)
	 *	paVecs: a list of parent vectors
	 *	returns: a realization of the node
	 */
	static nfg_Vec		Sample_movespace ( const nfg_Vecs& paVecs );
	/**
	 * Sample_level0: Sample the level 0 distribution of a decision node 
	 *	(level 0 distribution)
	 *	paVecs: a list of parent vectors
	 *	returns: a realization of the node
	 */
	static nfg_Vec		Sample_level0 ( const nfg_Vecs& paVecs );
	/**
	 * Partition_function_C: Specifies the partition function of node C
	 *	(used in branching to specify which node goes next)
	 *	vec: a realization of node C
	 *	returns: the child node to branch to
	 */
	static nfg_Vertex	Partition_function_C ( const nfg_Vec& vec );
	/**
	 * Dno_function_AliceFirst: Function that determines whether nodes is dno
	 *		for nodes in the Alice goes first branch.
	 *	optInstVecs - instantiation of the net so far
	 *	returns: true if dno, false otherwise
	 */
	static bool Dno_function_AliceFirst ( const nfg_OptInst& optInstVecs );
	/**
	 * Dno_function_BobFirst: Function that determines whether nodes is dno
	 *		for nodes in the Bob goes first branch.
	 *	optInstVecs - instantiation of the net so far
	 *	returns: true if dno, false otherwise
	 */
	static bool Dno_function_BobFirst ( const nfg_OptInst& optInstVecs );
	/**
	 * Analysis1: Performs a series of analyses based on a bnet
	 *	bnet: nfg object
	 */
	static void			Analysis1 ( BayesNet& bnet );
	/**
	 * Analysis2: Performs a series of analyses based on a bnet
	 *	bnet: nfg object
	 */
	static void			Analysis2 ( BayesNet& bnet );
};

#endif

