
// *************************************************************************
//	LibNFG: Net-form game library
//
//	Author:		Ritchie Lee
//
//  Copyright 2011 United States Government as represented by the Administrator 
//  of the National Aeronautics and Space Administration.  All Rights Reserved.  
//  Licensed under the Apache License, Version 2.0 (the "License"); you may not 
//  use this file except in compliance with the License.  You may obtain a copy 
//  of the License at http://www.apache.org/licenses/LICENSE-2.0
//  Unless required by applicable law or agreed to in writing, software 
//  distributed under the License is distributed on an "AS IS" BASIS, WITHOUT 
//  WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the 
//  License for the specific language governing permissions and limitations 
//  under the License.
//
//	Revision Log
//
//	Username	Date		Description
//	-------------------------------------------------------
//	rlee		20111001	Created
//
// *************************************************************************

#include "net_hitlcollisionavoid2.h"

#include <boost/lambda/bind.hpp>
#include <boost/random.hpp>
#include <boost/math/distributions/normal.hpp>
#include <boost/numeric/ublas/vector.hpp>
#include <boost/numeric/ublas/matrix.hpp>
#include <boost/numeric/ublas/operation.hpp>
#include <boost/algorithm/string.hpp>
#include <boost/lexical_cast.hpp>
#include <string>
#include <iostream>
#include "bayesnet.h"
#include "alg_levelkrelaxed.h"
#include "pointmasskinematics.h"
#include "cpd.h"
#include "cpd_gaussian.h"
#include "cpd_decision.h"
#include "player.h"
#include "if_matlab.h"
#include "if_bayesnettoolbox.h"
#include "if_mpi.h"

using namespace nfg;
using namespace std;
using namespace boost;
using namespace boost::lambda;
using namespace boost::math;
using namespace boost::numeric;

// *************************************************************************
// *************************************************************************
net_HitlCollisionAvoid2::net_HitlCollisionAvoid2 ( int argc, char *argv[] )
{
	// Defaults
	int numsamp(100); 
	double bias_x(300); 
	double bias_y(300);
	double  utilweight(0.8);
	int     klevel(1);    
	int     mvalue(5);
	int     mpvalue(10);
	string encounterFile ("");

	// Extract parameters from arguments
	numsamp = atoi(argv[2]);
	bias_x = atof(argv[3]);
	bias_y = atof(argv[4]);
	utilweight = atof(argv[5]);
	klevel = atoi(argv[6]);
	mvalue = atoi(argv[7]);
	mpvalue = atoi(argv[8]);
	encounterFile = argv[9];
	
	// Pass in parameters to be passed in from external, such as Matlab

	// Try opening encounterFile if one was specified 
	if( encounterFile != "" )
	{
		m_encFile.open( encounterFile.c_str(), ifstream::in );

		if ( m_encFile.is_open() )
			cout << "net_HitlCollisionAvoid2: Encounter file was opened successfully." << endl;
	}

	// Set parameters here
	m_PARAMS.m_SIGMA_OW_ESTLOC_X	= 900; // sigma_ow = 900; sigma_in = 600
	m_PARAMS.m_SIGMA_OW_ESTLOC_Y	= 900;
	m_PARAMS.m_SIGMA_OW_ESTLOC_Z	= 0.0;
	m_PARAMS.m_SIGMA_OW_ESTVEL_X	= 318; // estimated from (Graf, Warren, & Maloney, 2005)
	m_PARAMS.m_SIGMA_OW_ESTVEL_Y	= 318; // estimated from (Graf, Warren, & Maloney, 2005)
	m_PARAMS.m_SIGMA_OW_ESTVEL_Z	= 0.0;
	m_PARAMS.m_SIGMA_OW_ESTACC_X	= 0.0;
	m_PARAMS.m_SIGMA_OW_ESTACC_Y	= 0.0;
	m_PARAMS.m_SIGMA_OW_ESTACC_Z	= 0.0;
	m_PARAMS.m_SIGMA_IN_ESTLOC_X	= 600; // sigma_ow = 1800; sigma_in = 900; sigma_c = 600
	m_PARAMS.m_SIGMA_IN_ESTLOC_Y	= 600;
	m_PARAMS.m_SIGMA_IN_ESTLOC_Z	= 0.0;
	m_PARAMS.m_SIGMA_IN_ESTVEL_X	= 318; // estimated from (Graf, Warren, & Maloney, 2005)
	m_PARAMS.m_SIGMA_IN_ESTVEL_Y	= 318; // estimated from (Graf, Warren, & Maloney, 2005)
	m_PARAMS.m_SIGMA_IN_ESTVEL_Z	= 0.0;
	m_PARAMS.m_SIGMA_IN_ESTACC_X	= 0.0;
	m_PARAMS.m_SIGMA_IN_ESTACC_Y	= 0.0;
	m_PARAMS.m_SIGMA_IN_ESTACC_Z	= 0.0;
	m_PARAMS.m_BIAS_OW_ESTLOC_X		= bias_x;
	m_PARAMS.m_BIAS_OW_ESTLOC_Y		= bias_y;
	m_PARAMS.m_BIAS_OW_ESTLOC_Z		= 0.0;
	m_PARAMS.m_BIAS_OW_ESTVEL_X		= 0;
	m_PARAMS.m_BIAS_OW_ESTVEL_Y		= 0;
	m_PARAMS.m_BIAS_OW_ESTVEL_Z		= 0.0;
	m_PARAMS.m_BIAS_OW_ESTACC_X		= 0.0;
	m_PARAMS.m_BIAS_OW_ESTACC_Y		= 0.0;
	m_PARAMS.m_BIAS_OW_ESTACC_Z		= 0.0;
	m_PARAMS.m_BIAS_IN_ESTLOC_X		= 0;
	m_PARAMS.m_BIAS_IN_ESTLOC_Y		= 0;
	m_PARAMS.m_BIAS_IN_ESTLOC_Z		= 0.0;
	m_PARAMS.m_BIAS_IN_ESTVEL_X		= 0;
	m_PARAMS.m_BIAS_IN_ESTVEL_Y		= 0; 
	m_PARAMS.m_BIAS_IN_ESTVEL_Z		= 0.0;
	m_PARAMS.m_BIAS_IN_ESTACC_X		= 0.0;
	m_PARAMS.m_BIAS_IN_ESTACC_Y		= 0.0;
	m_PARAMS.m_BIAS_IN_ESTACC_Z		= 0.0;
	m_PARAMS.m_VIEWTIME				= 5.0;
	m_PARAMS.m_TIMESTEPSIZE		= .01;
	m_PARAMS.m_MAXANGLE			= 1.0;	// radians
	m_PARAMS.m_SIMTIME			= 5.0;	// seconds
	m_PARAMS.m_TIME2COLLIDE		= 20;	// seconds
	m_PARAMS.m_ALTITUDE			= 35000; // feet
	m_PARAMS.m_P1SPEED			= 800;	// feet per second
	m_PARAMS.m_P2SPEED			= 750;	// feet per second
	m_PARAMS.m_FOV				= 0.785398163; // 45 degrees in radians
	m_PARAMS.m_P2ANGLEMINMAX		= 0.0872664626; // 5 degrees in radians
	m_PARAMS.m_NUMSAMPLESPERPROC	= numsamp;
	m_PARAMS.m_KLEVEL			= klevel;
	m_PARAMS.m_MVALUE			= mvalue;
	m_PARAMS.m_MPVALUE			= mpvalue;
	m_PARAMS.m_UTILWGHT			= utilweight;
}
// *************************************************************************
// *************************************************************************
net_HitlCollisionAvoid2::~net_HitlCollisionAvoid2 ( void )
{
	if( m_encFile.is_open() )
		m_encFile.close();
}
// *************************************************************************
// *************************************************************************
void net_HitlCollisionAvoid2::Run ( void )
{
	//*********************************************
	///	 Set up Bayes Net

	// Specify vertex names
	string vertexNames[] = 
	{	"S1I", "S2I", "S1", "S2", "O1OW", "O1IN", "O2OW", "O2IN", "A1", "A2", "S3", "S4"  };	

	// Define Edges of the graph
	nfg_Edge edgeArray[] = 
	{ 
		nfg_Edge(S1I,S1), nfg_Edge(S2I,S2), 
		nfg_Edge(S1,O2OW), nfg_Edge(S2,O1OW),
		nfg_Edge(S1,O2IN), nfg_Edge(S2,O1IN),
		nfg_Edge(S1,A1), nfg_Edge(S2,A2),
		nfg_Edge(O1OW,A1), nfg_Edge(O2OW,A2),
		nfg_Edge(O1IN,A1), nfg_Edge(O2IN,A2),
		nfg_Edge(S1,S3), nfg_Edge(S2,S4),
		nfg_Edge(A1,S3), nfg_Edge(A2,S4)
		
	};

	int dimArray[] =
	{ 9, 9, 9, 9, 9, 9, 9, 9, 1, 1, 9, 9 };

	// Specify the space of each vertex
	nfg_Spaces spaces(N);
	spaces.at(A1).push_back( nfg_Space1D(1, nfg_Range(-m_PARAMS.m_MAXANGLE,
						m_PARAMS.m_MAXANGLE) ));
	spaces.at(A2).push_back( nfg_Space1D(1, nfg_Range(-m_PARAMS.m_MAXANGLE,
						m_PARAMS.m_MAXANGLE) ));

	// Note: no constraints on all other nodes because they are continuous

	// Specify vertex type
	nfg_eVertexType vertexTypes[] =
	{	CHANCE, CHANCE, CHANCE, CHANCE, CHANCE, CHANCE, CHANCE, CHANCE, DECISION, DECISION,
			CHANCE, CHANCE	};

	////	Make Bayes Net
	// Pass all specified values above to the bnet constructor
	BayesNet bnet( "HITL Collision Avoid 2", N,
		nfg_Edges(edgeArray,edgeArray+sizeof(edgeArray)/sizeof(nfg_Edge)), 
		nfg_Dim( dimArray,dimArray+sizeof(dimArray)/sizeof(int)),
		spaces,
		nfg_Mask( N, true ),	// all nodes continuous
		nfg_eVertexTypes(vertexTypes,vertexTypes+N),
		nfg_Mask(N, false),		// no branch nodes
		nfg_Mask(N, false));	// no nodes with dno
	// Set vertex names
	bnet.Set_vertex_names ( vector<string>(vertexNames,vertexNames+N) );

	//	Declare players
	Player P1 ( bind(&net_HitlCollisionAvoid2::Util_function_P1, this,_1) );	// pass in utility function
	Player P2 ( bind(&net_HitlCollisionAvoid2::Util_function_P2, this, _1) );	// pass in utility function
	// Specify which decision nodes are controlled by which players
	P1.Add_controlled_node(A1);
	P2.Add_controlled_node(A2);

	// Associate players with bnet
	bnet.Add_player(P1);
	bnet.Add_player(P2);

	//	Use library cpds where available
	// O1 observes S2
	// Set variances
	nfg_Vec O1OW_sigArray;
	O1OW_sigArray.push_back( m_PARAMS.m_SIGMA_OW_ESTLOC_X );
	O1OW_sigArray.push_back( m_PARAMS.m_SIGMA_OW_ESTLOC_Y );
	O1OW_sigArray.push_back( m_PARAMS.m_SIGMA_OW_ESTLOC_Z );
	O1OW_sigArray.push_back( m_PARAMS.m_SIGMA_OW_ESTVEL_X );
	O1OW_sigArray.push_back( m_PARAMS.m_SIGMA_OW_ESTVEL_Y );
	O1OW_sigArray.push_back( m_PARAMS.m_SIGMA_OW_ESTVEL_Z );
	O1OW_sigArray.push_back( m_PARAMS.m_SIGMA_OW_ESTACC_X );
	O1OW_sigArray.push_back( m_PARAMS.m_SIGMA_OW_ESTACC_Y );
	O1OW_sigArray.push_back( m_PARAMS.m_SIGMA_OW_ESTACC_Z );
	nfg_Vec O1OW_biasArray;
	O1OW_biasArray.push_back( m_PARAMS.m_BIAS_OW_ESTLOC_X );
	O1OW_biasArray.push_back( m_PARAMS.m_BIAS_OW_ESTLOC_Y );
	O1OW_biasArray.push_back( m_PARAMS.m_BIAS_OW_ESTLOC_Z );
	O1OW_biasArray.push_back( m_PARAMS.m_BIAS_OW_ESTVEL_X );
	O1OW_biasArray.push_back( m_PARAMS.m_BIAS_OW_ESTVEL_Y );
	O1OW_biasArray.push_back( m_PARAMS.m_BIAS_OW_ESTVEL_Z );
	O1OW_biasArray.push_back( m_PARAMS.m_BIAS_OW_ESTACC_X );
	O1OW_biasArray.push_back( m_PARAMS.m_BIAS_OW_ESTACC_Y );
	O1OW_biasArray.push_back( m_PARAMS.m_BIAS_OW_ESTACC_Z );
	cpd_Gaussian	O1OW_cpd ( O1OW_sigArray, O1OW_biasArray );

	nfg_Vec O1IN_sigArray;
	O1IN_sigArray.push_back( m_PARAMS.m_SIGMA_IN_ESTLOC_X );
	O1IN_sigArray.push_back( m_PARAMS.m_SIGMA_IN_ESTLOC_Y );
	O1IN_sigArray.push_back( m_PARAMS.m_SIGMA_IN_ESTLOC_Z );
	O1IN_sigArray.push_back( m_PARAMS.m_SIGMA_IN_ESTVEL_X );
	O1IN_sigArray.push_back( m_PARAMS.m_SIGMA_IN_ESTVEL_Y );
	O1IN_sigArray.push_back( m_PARAMS.m_SIGMA_IN_ESTVEL_Z );
	O1IN_sigArray.push_back( m_PARAMS.m_SIGMA_IN_ESTACC_X );
	O1IN_sigArray.push_back( m_PARAMS.m_SIGMA_IN_ESTACC_Y );
	O1IN_sigArray.push_back( m_PARAMS.m_SIGMA_IN_ESTACC_Z );
	nfg_Vec O1IN_biasArray;
	O1IN_biasArray.push_back( m_PARAMS.m_BIAS_IN_ESTLOC_X );
	O1IN_biasArray.push_back( m_PARAMS.m_BIAS_IN_ESTLOC_Y );
	O1IN_biasArray.push_back( m_PARAMS.m_BIAS_IN_ESTLOC_Z );
	O1IN_biasArray.push_back( m_PARAMS.m_BIAS_IN_ESTVEL_X );
	O1IN_biasArray.push_back( m_PARAMS.m_BIAS_IN_ESTVEL_Y );
	O1IN_biasArray.push_back( m_PARAMS.m_BIAS_IN_ESTVEL_Z );
	O1IN_biasArray.push_back( m_PARAMS.m_BIAS_IN_ESTACC_X );
	O1IN_biasArray.push_back( m_PARAMS.m_BIAS_IN_ESTACC_Y );
	O1IN_biasArray.push_back( m_PARAMS.m_BIAS_IN_ESTACC_Z );
	cpd_Gaussian	O1IN_cpd ( O1IN_sigArray, O1IN_biasArray );
	
	// O2 observes S1
	// Set variances
	nfg_Vec O2OW_sigArray;
	O2OW_sigArray.push_back( m_PARAMS.m_SIGMA_OW_ESTLOC_X );
	O2OW_sigArray.push_back( m_PARAMS.m_SIGMA_OW_ESTLOC_Y );
	O2OW_sigArray.push_back( m_PARAMS.m_SIGMA_OW_ESTLOC_Z );
	O2OW_sigArray.push_back( m_PARAMS.m_SIGMA_OW_ESTVEL_X );
	O2OW_sigArray.push_back( m_PARAMS.m_SIGMA_OW_ESTVEL_Y );
	O2OW_sigArray.push_back( m_PARAMS.m_SIGMA_OW_ESTVEL_Z );
	O2OW_sigArray.push_back( m_PARAMS.m_SIGMA_OW_ESTACC_X );
	O2OW_sigArray.push_back( m_PARAMS.m_SIGMA_OW_ESTACC_Y );
	O2OW_sigArray.push_back( m_PARAMS.m_SIGMA_OW_ESTACC_Z );
	nfg_Vec O2OW_biasArray;
	O2OW_biasArray.push_back( m_PARAMS.m_BIAS_OW_ESTLOC_X );
	O2OW_biasArray.push_back( m_PARAMS.m_BIAS_OW_ESTLOC_Y );
	O2OW_biasArray.push_back( m_PARAMS.m_BIAS_OW_ESTLOC_Z );
	O2OW_biasArray.push_back( m_PARAMS.m_BIAS_OW_ESTVEL_X );
	O2OW_biasArray.push_back( m_PARAMS.m_BIAS_OW_ESTVEL_Y );
	O2OW_biasArray.push_back( m_PARAMS.m_BIAS_OW_ESTVEL_Z );
	O2OW_biasArray.push_back( m_PARAMS.m_BIAS_OW_ESTACC_X );
	O2OW_biasArray.push_back( m_PARAMS.m_BIAS_OW_ESTACC_Y );
	O2OW_biasArray.push_back( m_PARAMS.m_BIAS_OW_ESTACC_Z );
	cpd_Gaussian	O2OW_cpd ( O2OW_sigArray, O2OW_biasArray );

	nfg_Vec O2IN_sigArray;
	O2IN_sigArray.push_back( m_PARAMS.m_SIGMA_IN_ESTLOC_X );
	O2IN_sigArray.push_back( m_PARAMS.m_SIGMA_IN_ESTLOC_Y );
	O2IN_sigArray.push_back( m_PARAMS.m_SIGMA_IN_ESTLOC_Z );
	O2IN_sigArray.push_back( m_PARAMS.m_SIGMA_IN_ESTVEL_X );
	O2IN_sigArray.push_back( m_PARAMS.m_SIGMA_IN_ESTVEL_Y );
	O2IN_sigArray.push_back( m_PARAMS.m_SIGMA_IN_ESTVEL_Z );
	O2IN_sigArray.push_back( m_PARAMS.m_SIGMA_IN_ESTACC_X );
	O2IN_sigArray.push_back( m_PARAMS.m_SIGMA_IN_ESTACC_Y );
	O2IN_sigArray.push_back( m_PARAMS.m_SIGMA_IN_ESTACC_Z );
	nfg_Vec O2IN_biasArray;
	O2IN_biasArray.push_back( m_PARAMS.m_BIAS_IN_ESTLOC_X );
	O2IN_biasArray.push_back( m_PARAMS.m_BIAS_IN_ESTLOC_Y );
	O2IN_biasArray.push_back( m_PARAMS.m_BIAS_IN_ESTLOC_Z );
	O2IN_biasArray.push_back( m_PARAMS.m_BIAS_IN_ESTVEL_X );
	O2IN_biasArray.push_back( m_PARAMS.m_BIAS_IN_ESTVEL_Y );
	O2IN_biasArray.push_back( m_PARAMS.m_BIAS_IN_ESTVEL_Z );
	O2IN_biasArray.push_back( m_PARAMS.m_BIAS_IN_ESTACC_X );
	O2IN_biasArray.push_back( m_PARAMS.m_BIAS_IN_ESTACC_Y );
	O2IN_biasArray.push_back( m_PARAMS.m_BIAS_IN_ESTACC_Z );
	cpd_Gaussian	O2IN_cpd ( O2IN_sigArray, O2IN_biasArray );
	
	// Decision nodes
	cpd_Decision	A1_cpd ( P1 );	// associate with player
	cpd_Decision	A2_cpd ( P2 );	// associate with player

	// State nodes
	cpd_S1I		S1I_cpd( m_PARAMS );
	cpd_S2I		S2I_cpd( m_PARAMS );
	cpd_S12		S1_cpd( m_PARAMS );
	cpd_S12		S2_cpd( m_PARAMS );
	cpd_S34		S3_cpd( m_PARAMS );
	cpd_S34		S4_cpd( m_PARAMS );

	//	Link cpds to bnet
	bnet.Set_cpd(S1I, &S1I_cpd);
	bnet.Set_cpd(S2I, &S2I_cpd);
	bnet.Set_cpd(S1, &S1_cpd);
	bnet.Set_cpd(S2, &S2_cpd);
	bnet.Set_cpd(S3, &S3_cpd );
	bnet.Set_cpd(S4, &S4_cpd );
	bnet.Set_cpd(O1OW, &O1OW_cpd );
	bnet.Set_cpd(O2OW, &O2OW_cpd );
	bnet.Set_cpd(O1IN, &O1IN_cpd );
	bnet.Set_cpd(O2IN, &O2IN_cpd );
	bnet.Set_cpd(A1, &A1_cpd );
	bnet.Set_cpd(A2, &A2_cpd );

	//	Error check
	bassert ( bnet.Check() );

	//**************************************************
	///	Run main program
	
	// Draw graph (to make sure the code understood your graph correctly)
	if_BayesNetToolbox::Draw_graph( "HCA2_draw.m", bnet );

	// Run custom analysis on bnet
	Analysis1(bnet);

	return;
}
// *************************************************************************
// *************************************************************************
double net_HitlCollisionAvoid2::Util_function_P1 ( const nfg_Inst& instVecs )
{
	//	weights
	double a1 = m_PARAMS.m_UTILWGHT;
	double a2 = 1 - a1;

	//	Term 1: Distance to the other player, higher is better
	double u1 = sqrt ( pow(instVecs.at(S4).at(0)-instVecs.at(S3).at(0),2)+pow(instVecs.at(S4).at(1)-instVecs.at(S3).at(1),2) ) 
						/ ((m_PARAMS.m_P1SPEED + m_PARAMS.m_P2SPEED) * m_PARAMS.m_TIME2COLLIDE );

	//	Term 2: How much effort was spent to move, lower is better
	double u2 = abs ( instVecs.at(A1).at(0) / m_PARAMS.m_MAXANGLE );	// absolute angle of input

	//	Term 3: Right of way cost, to be defined...

	// linear combo of all terms
	return a1*u1 - a2*u2;
}
// *************************************************************************
// *************************************************************************
double net_HitlCollisionAvoid2::Util_function_P2 ( const nfg_Inst& instVecs )
{
	// weights
	double a1 = m_PARAMS.m_UTILWGHT;
	double a2 = 1 - a1;

	//	Term 1: Distance to the other player, higher is better
	double u1 = sqrt ( pow(instVecs.at(S4).at(0)-instVecs.at(S3).at(0),2)+pow(instVecs.at(S4).at(1)-instVecs.at(S3).at(1),2) ) 
						/ ((m_PARAMS.m_P1SPEED + m_PARAMS.m_P2SPEED) * m_PARAMS.m_TIME2COLLIDE );

	//	Term 2: How much effort was spent to move, lower is better
	double u2 = abs ( instVecs.at(A2).at(0) / m_PARAMS.m_MAXANGLE );	// absolute angle of input

	//	Term 3: Right of way cost, to be defined...

	// linear combo of all terms
	return a1*u1 - a2*u2;
}
// *************************************************************************
// *************************************************************************
nfg_Vec net_HitlCollisionAvoid2::Sample_movespace ( const nfg_Vecs& paVecs )
{
	//	The parent vectors are ignored

	// Uniform over space of possible angles (radians)
	return nfg_Vec(1, Rand_uniform_range(-m_PARAMS.m_MAXANGLE,
										m_PARAMS.m_MAXANGLE) );
}
// *************************************************************************
// *************************************************************************
nfg_Vec net_HitlCollisionAvoid2::Sample_level0 ( const nfg_Vecs& paVecs )
{
	//	The parent vectors are ignored

	// Uniform over space of possible angles (radians)
	return nfg_Vec(1, Rand_uniform_range(-m_PARAMS.m_MAXANGLE,
									m_PARAMS.m_MAXANGLE) );
}
// *************************************************************************
// *************************************************************************
void net_HitlCollisionAvoid2::Analysis1 ( BayesNet& bnet )
{
	// Declare inference engine
	alg_LevelKRelaxed engine;

	// Specify level0 distributions (functions) for each decision node
	nfg_SampleFuncs fo_Sample_level0_inst(N);
	fo_Sample_level0_inst.at(A1) = bind(&net_HitlCollisionAvoid2::Sample_level0, this, _1);	// specify level0 distr.
	fo_Sample_level0_inst.at(A2) = bind(&net_HitlCollisionAvoid2::Sample_level0, this, _1);	// specify level0 distr.

	// Specify satisficing distributions (functions) for each decision node
	nfg_SampleFuncs fo_Sample_satisf_inst(N);
	fo_Sample_satisf_inst.at(A1) = bind(&net_HitlCollisionAvoid2::Sample_movespace,this,_1);	// specify satisficing distr.
	fo_Sample_satisf_inst.at(A2) = bind(&net_HitlCollisionAvoid2::Sample_movespace,this,_1);	// specify satisficing distr.

	// Set level K parameters
	engine.Set_params( bnet, alg_LevelKRelaxed::D_RELAXED, 5, 10, fo_Sample_level0_inst,
								fo_Sample_satisf_inst,
								alg_LevelKRelaxed::BEST, true );
	
	// Collect some samples from each processor using parallel processing module
	if_Mpi::Parallel_exec( 
		bind( &net_HitlCollisionAvoid2::Analysis1_parProc, this, ref(engine), ref(bnet) ), 
		bind( &net_HitlCollisionAvoid2::Analysis1_postProc,this,_1) );
	
}
// *************************************************************************
// *************************************************************************
nfg_WtdInsts net_HitlCollisionAvoid2::Analysis1_parProc ( alg_LevelKRelaxed& engine, BayesNet& bnet )
{
	// Used to contain the samples
	nfg_WtdInsts wInsts;

	// Collect some samples of the net
	for ( int i = 0; i < m_PARAMS.m_NUMSAMPLESPERPROC; i++ )
	{
		cout << i << endl;		// Progress indicator to console

		// If an encounter file is used, parse the next line
		if( m_encFile.is_open() )
		{
			// If file read is good, read a line
			if ( m_encFile.good() )
			{
				string line;

				// Note: getline uses carriage return as token
				getline( m_encFile, line );
				
				vector<string> tokVec; // To contain parsed tokens
				// parse using Boost string algorithm library
				split( tokVec, line, is_any_of(","), token_compress_on );
				bassert( tokVec.size() == 2 );	// Quick check to make sure parsing is valid
				
				// Extract to correct variables and variable types using Boost function
				int config = lexical_cast<int>( tokVec.at(0) );
				double phi = lexical_cast<double>( tokVec.at(1) );

				//cout << "config: " << config << ", phi:" << phi << endl;
				
				nfg_OptInst evidence( N );
				// Using the parameters, generate what S2I should be, set it as evidence
				evidence.at(S2I) = Generate_deterministic_S2I( config, phi );

				// Generate a sample and add it to the list
				nfg_OptWtdInst owInst = engine.LW_forward_sample_detailed( bnet, evidence, m_PARAMS.m_KLEVEL,
						nfg_Mask(), nfg_OptWtdInst(), nfg_PropFuncInst(), nfg_PropProbFuncInst());
				wInsts.push_back( 
					nfg_WtdInst( Convert_OptInst_to_Inst( owInst.m_optInst ), 1 ) );
			}
			// If file read was not good, throw an error
			else
			{	
				cout << "Error: net_HitlCollisionAvoid2::Analysis1_parProc: Encounter file not valid!" << endl;
				bassert( false );
			}
		}
		// If no encounter file, proceed as usual with no evidence specified
		else
		{
			// Generate a sample and add it to the list
			wInsts.push_back( engine.LW_forward_sample( bnet, m_PARAMS.m_KLEVEL ) );
		}
	}

	return wInsts;
}
// *************************************************************************
// *************************************************************************
void net_HitlCollisionAvoid2::Analysis1_postProc ( const vector<nfg_WtdInsts>& wInstsVec  )
{
	nfg_WtdInsts finalList;

	// Unpack the lists and reconstruct a single large list
	foreach( const nfg_WtdInsts& wInsts, wInstsVec )
	{
		finalList.insert( finalList.end(), wInsts.begin(), wInsts.end() );
	}

	// Output the samples to a Matlab file for post-processing
	if_Matlab::Export_WtdInsts_to_file ( "HCA2_vars.m", "samples", finalList, false );

	//	Track players' utility distribution
	nfg_Vec P1Utils, P2Utils;
	foreach ( nfg_WtdInst wInst, finalList )
	{
		// Map each instantiation (sample) of the net to the player's utilities
		// and save them to a list
		P1Utils.push_back( Util_function_P1(wInst.m_inst) );
		P2Utils.push_back( Util_function_P2(wInst.m_inst) );
	}
	// Output the utility values to a Matlab file for post-processing
	if_Matlab::Export_Vec_to_file ( "HCA2_vars.m", "P1Utils", P1Utils, true );
	if_Matlab::Export_Vec_to_file ( "HCA2_vars.m", "P2Utils", P2Utils, true );

}
// *************************************************************************
// *************************************************************************
nfg_Vec net_HitlCollisionAvoid2::Generate_deterministic_S2I ( const int config, const double phi )
{
	// Player 2's encounter model (see accompanying documentation)
	// Pc is intended point of collision
	// Pm is the point of shortest distance between Pc and line P1 P2
	bassert( config >= -2 && config <= 2 );	// Check that it is in range
	double psi		= (double)config * m_PARAMS.m_FOV / 4;	// angle between d3 and d1
	double d1		= m_PARAMS.m_P1SPEED * m_PARAMS.m_TIME2COLLIDE;	// distance P1 to Pc
	double d2		= d1 * sin( abs(psi) );			// distance Pm to Pc
	double d3		= d1 * cos( abs(psi) );			// distance P1 to Pm
	double d5		= m_PARAMS.m_P2SPEED * m_PARAMS.m_TIME2COLLIDE;	// distance P2 to Pc
	double theta	= asin( d2 / d5 );		// angle between d4 and d5
	double d4		= d5 / cos( theta );	// distance Pm to P2
	double r		= d3 + d4;				// distance P1 to P2

	// Calculate position
	ublas::vector<double> P1(2);
	P1(0) = 0;				// P1x
	P1(1) = 0;				// P1y

	// Player 1's V
	ublas::vector<double> V1(2);
	V1(0) = 0;				// V1x
	V1(1) = m_PARAMS.m_P1SPEED;		// V1y

	// unit vector V1
	ublas::vector<double> V1hat(2);
	V1hat = V1 / norm_2(V1);

	// Define rotation matrix (only rotate in the horizontal plane)
	ublas::matrix<double> R1(2,2);
	R1(0,0) = cos( psi );
	R1(0,1) = -sin( psi );
	R1(1,0) = sin( psi );
	R1(1,1) = cos( psi );

	// Do multiplication to rotate vector
	ublas::vector<double> V1hatNew(2);
	ublas::axpy_prod(R1, V1hat, V1hatNew, true);  // V1new = R1 * V1

	// Use solved r from above to determine P2
	ublas::vector<double> P2(2);
	P2 = P1 + r * V1hatNew;
	
	// Point of collision
	ublas::vector<double> Pc(2);
	Pc		= P1 + m_PARAMS.m_TIME2COLLIDE * V1;

	// Calculate velocity to get to collision point
	ublas::vector<double> V2(2);
	V2		= (Pc - P2) / m_PARAMS.m_TIME2COLLIDE;

	// Rotate the velocity vector slightly away randomly
	// Define rotation matrix (only rotate in the horizontal plane)
	bassert( phi >= -m_PARAMS.m_P2ANGLEMINMAX && phi <= m_PARAMS.m_P2ANGLEMINMAX );	// check range of phi
	ublas::matrix<double> R2(2,2);
	R2(0,0) = cos( phi );
	R2(0,1) = -sin( phi );
	R2(1,0) = sin( phi );
	R2(1,1) = cos( phi );

	// Do multiplication to rotate vector
	ublas::vector<double> V2rand(2);
	ublas::axpy_prod(R2, V2, V2rand, true);  // V2rand = R2 * V2

	nfg_Vec P2vec;
	// Set position
	P2vec.push_back(P2(0));
	P2vec.push_back(P2(1));
	P2vec.push_back(m_PARAMS.m_ALTITUDE);
	// Set velocity
	P2vec.push_back(V2rand(0));
	P2vec.push_back(V2rand(1));
	P2vec.push_back(0);
	// Set acceleration
	P2vec.push_back(0);
	P2vec.push_back(0);
	P2vec.push_back(0);

	return P2vec;
}
// *************************************************************************
// *************************************************************************
void net_HitlCollisionAvoid2::Unit_subtest_file_read( const string filename )
{
	string line;
	ifstream myfile( filename.c_str() );

	if (myfile.is_open())
	{
		while ( myfile.good() )
		{
			getline( myfile, line );
			cout << line << endl;
			
			vector<string> tokVec; // Search for tokens
			split( tokVec, line, is_any_of(","), token_compress_on ); // tokVec = split vec
			bassert( !tokVec.empty() );
			
			
			foreach( string& s, tokVec )
			{
				cout << s << endl;

				// use boost to convert string number
				cout << lexical_cast<double>(s) << endl;
			}
		}
		myfile.close();
	}
	else 
	{
		cout << "net_HitlCollisionAvoid2::Unit_subtest_file_read: Unable to open file";
		bassert( false );
	}


	return;
}


/**
 * Define CPD class functions
 */
// *************************************************************************
// *************************************************************************
nfg_Vec net_HitlCollisionAvoid2::cpd_S1I::Sample_node ( const nfg_Vecs& paVecs )
{
	// flying straight forward, no accel
	nfg_Vec outVec(9,0);
	outVec.at(2) = m_PARAMS.m_ALTITUDE;		// ft
	outVec.at(4) = m_PARAMS.m_P1SPEED;		// ft/s along y-axis

	// everything else is 0...

	return outVec;
}
// *************************************************************************
// *************************************************************************
double net_HitlCollisionAvoid2::cpd_S1I::Get_probability_prop ( const nfg_Vec& outVec,
														const nfg_Vecs& paVecs )
{
	// Deterministic encounter model for player 1, so
	nfg_Vec outVec2 = Sample_node( paVecs );
	
	// If vectors are equal, return P=1, otherwise P=0
	return ( Is_equal(outVec,outVec2) ? 1.0:0.0 );
}
// *************************************************************************
// *************************************************************************
nfg_Vec net_HitlCollisionAvoid2::cpd_S2I::Sample_node ( const nfg_Vecs& paVecs )
{
	// Player 2's encounter model (see accompanying documentation)
	// Pc is intended point of collision
	// Pm is the point of shortest distance between Pc and line P1 P2
	int config		= Rand_int(-2,2);				// randomly choose from 5 configurations
	double psi		= (double)config * m_PARAMS.m_FOV / 4;	// angle between d3 and d1
	double d1		= m_PARAMS.m_P1SPEED * m_PARAMS.m_TIME2COLLIDE;	// distance P1 to Pc
	double d2		= d1 * sin( abs(psi) );			// distance Pm to Pc
	double d3		= d1 * cos( abs(psi) );			// distance P1 to Pm
	double d5		= m_PARAMS.m_P2SPEED * m_PARAMS.m_TIME2COLLIDE;	// distance P2 to Pc
	double theta	= asin( d2 / d5 );		// angle between d4 and d5
	double d4		= d5 / cos( theta );	// distance Pm to P2
	double r		= d3 + d4;				// distance P1 to P2

	// Calculate position
	ublas::vector<double> P1(2);
	P1(0) = 0;				// P1x
	P1(1) = 0;				// P1y

	// Player 1's V
	ublas::vector<double> V1(2);
	V1(0) = 0;				// V1x
	V1(1) = m_PARAMS.m_P1SPEED;		// V1y

	// unit vector V1
	ublas::vector<double> V1hat(2);
	V1hat = V1 / norm_2(V1);

	// Define rotation matrix (only rotate in the horizontal plane)
	ublas::matrix<double> R1(2,2);
	R1(0,0) = cos( psi );
	R1(0,1) = -sin( psi );
	R1(1,0) = sin( psi );
	R1(1,1) = cos( psi );

	// Do multiplication to rotate vector
	ublas::vector<double> V1hatNew(2);
	ublas::axpy_prod(R1, V1hat, V1hatNew, true);  // V1new = R1 * V1

	// Use solved r from above to determine P2
	ublas::vector<double> P2(2);
	P2 = P1 + r * V1hatNew;
	
	// Point of collision
	ublas::vector<double> Pc(2);
	Pc		= P1 + m_PARAMS.m_TIME2COLLIDE * V1;

	// Calculate velocity to get to collision point
	ublas::vector<double> V2(2);
	V2		= (Pc - P2) / m_PARAMS.m_TIME2COLLIDE;

	// Rotate the velocity vector slightly away randomly
	// Define rotation matrix (only rotate in the horizontal plane)
	double phi = Rand_uniform_range( -m_PARAMS.m_P2ANGLEMINMAX, 
								m_PARAMS.m_P2ANGLEMINMAX );
	ublas::matrix<double> R2(2,2);
	R2(0,0) = cos( phi );
	R2(0,1) = -sin( phi );
	R2(1,0) = sin( phi );
	R2(1,1) = cos( phi );

	// Do multiplication to rotate vector
	ublas::vector<double> V2rand(2);
	ublas::axpy_prod(R2, V2, V2rand, true);  // V2rand = R2 * V2

	nfg_Vec P2vec;
	// Set position
	P2vec.push_back(P2(0));
	P2vec.push_back(P2(1));
	P2vec.push_back(m_PARAMS.m_ALTITUDE);
	// Set velocity
	P2vec.push_back(V2rand(0));
	P2vec.push_back(V2rand(1));
	P2vec.push_back(0);
	// Set acceleration
	P2vec.push_back(0);
	P2vec.push_back(0);
	P2vec.push_back(0);

	return P2vec;
}
// *************************************************************************
// *************************************************************************
double net_HitlCollisionAvoid2::cpd_S2I::Get_probability_prop ( const nfg_Vec& outVec,
														const nfg_Vecs& paVecs )
{
	double prob = 1/5;	// uniform over 5 possible starting positions
	prob *= 1/(2*m_PARAMS.m_P2ANGLEMINMAX); // uniform over angle

	return prob;
}
// *************************************************************************
// *************************************************************************
nfg_Vec net_HitlCollisionAvoid2::cpd_S12::Sample_node ( const nfg_Vecs& paVecs )
{
	int numSteps = (int)ceil(m_PARAMS.m_VIEWTIME/m_PARAMS.m_TIMESTEPSIZE);

	nfg_Vec S = paVecs.at(0);	// incoming state

	// Use point mass physics class
	PointMassKinematics mass( S.at(0), S.at(1), S.at(2),
							S.at(3), S.at(4), S.at(5) );
	mass.Set_params( m_PARAMS.m_TIMESTEPSIZE );
	
	// Simulate forward a preset number of steps
	for ( int i = 0; i < numSteps; i++ )
	{
		mass.Step ( S.at(6), S.at(7), S.at(8) );
	}

	// Return the final state
	nfg_Vec outVec;
	outVec.push_back( mass.Get_x() );
	outVec.push_back( mass.Get_y() );
	outVec.push_back( mass.Get_z() );
	outVec.push_back( mass.Get_vx() );
	outVec.push_back( mass.Get_vy() );
	outVec.push_back( mass.Get_vz() );

	// Constant accel
	outVec.push_back( S.at(6) );
	outVec.push_back( S.at(7) );
	outVec.push_back( S.at(8) );

	return outVec;
}
// *************************************************************************
// *************************************************************************
double net_HitlCollisionAvoid2::cpd_S12::Get_probability_prop ( const nfg_Vec& outVec,
															const nfg_Vecs& paVecs )
{
	// S1I and S2I are uniform, so S1 and S2 should be also.
	return 1.0;
}
// *************************************************************************
// *************************************************************************
nfg_Vec net_HitlCollisionAvoid2::cpd_S34::Sample_node ( const nfg_Vecs& paVecs )
{
	// Simulation parameters
	const int numSteps = (int)ceil( m_PARAMS.m_SIMTIME/m_PARAMS.m_TIMESTEPSIZE ); 
	
	nfg_Vec S = paVecs.at(0);		// incoming state
	double angle = paVecs.at(1).at(0);	// action by pilot (angle to rotate)

	// Apply impulse to velocity vector to rotate it in direction of turn
	// and override the current state
	// Assemble velocity vector
	ublas::vector<double> v(2);
	v(0) = S.at(3);		// vx
	v(1) = S.at(4);		// vy

	// Define rotation matrix (only rotate in the horizontal plane)
	ublas::matrix<double> R(2,2);
	R(0,0) = cos( angle );
	R(0,1) = -sin( angle );
	R(1,0) = sin( angle );
	R(1,1) = cos( angle );

	// Do multiplication to rotate v vector
	ublas::vector<double> vnew(2);
	ublas::axpy_prod(R, v, vnew, true);  // vnew = R * v

	// Use point mass physics class and overwrite current velocity state
	PointMassKinematics mass( S.at(0), S.at(1), S.at(2), vnew(0), vnew(1), S.at(5) );
	mass.Set_params( m_PARAMS.m_TIMESTEPSIZE );
	
	// Simulate forward a preset number of steps
	for ( int i = 0; i < numSteps; i++ )
	{
		mass.Step ( S.at(6), S.at(7), S.at(8) );
	}

	// Return the final state
	nfg_Vec outVec;
	outVec.push_back( mass.Get_x() );
	outVec.push_back( mass.Get_y() );
	outVec.push_back( mass.Get_z() );
	outVec.push_back( mass.Get_vx() );
	outVec.push_back( mass.Get_vy() );
	outVec.push_back( mass.Get_vz() );

	// Constant accel
	outVec.push_back( S.at(6) );
	outVec.push_back( S.at(7) );
	outVec.push_back( S.at(8) );

	return outVec;
}
// *************************************************************************
// *************************************************************************
double net_HitlCollisionAvoid2::cpd_S34::Get_probability_prop ( const nfg_Vec& outVec,
															const nfg_Vecs& paVecs )
{
	// Deterministic simulation
	nfg_Vec outVec2 = Sample_node( paVecs );
	
	// If vectors are equal, return P=1, otherwise P=0
	return ( Is_equal(outVec,outVec2) ? 1.0:0.0 );
}


