
// *************************************************************************
//	LibNFG: Net-form game library
//
//	Author:		Ritchie Lee
//
//  Copyright 2011 United States Government as represented by the Administrator 
//  of the National Aeronautics and Space Administration.  All Rights Reserved.  
//  Licensed under the Apache License, Version 2.0 (the "License"); you may not 
//  use this file except in compliance with the License.  You may obtain a copy 
//  of the License at http://www.apache.org/licenses/LICENSE-2.0
//  Unless required by applicable law or agreed to in writing, software 
//  distributed under the License is distributed on an "AS IS" BASIS, WITHOUT 
//  WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the 
//  License for the specific language governing permissions and limitations 
//  under the License.
//
//	Revision Log
//
//	Username	Date		Description
//	-------------------------------------------------------
//	rlee		20111001	Created
//
// *************************************************************************

#ifndef INCLUDED_NET_HITLCOLLISIONAVOID2_H
#define INCLUDED_NET_HITLCOLLISIONAVOID2_H

#include <fstream>
#include "helper.h"
#include "types.h"
#include "bayesnet.h"
#include "alg_levelkrelaxed.h"

/**
 * "HITL Collision Avoidance 2" example from Erik Schlicht (MIT-LL)
 *	In support of Human-in-the-loop experiments.  2 aircraft are approaching 
 *	a potential collision.  They see "movies" of varying lengths of the encounter 
 *	before being asked to make a one-shot decision to go left, straight, or right
 *	with stochastic action implications on their aircraft.
 *	Observation variance varies with length of movie presented.
 *	This version (net_HitlCollisionAvoid2) is the high-fidelity version with
 *	2 observation nodes per player.
 */
class net_HitlCollisionAvoid2
{
	// Make convenient labels for the vertices
	enum { S1I, S2I, S1, S2, O1OW, O1IN, O2OW, O2IN, A1, A2, S3, S4, N };

public:

	// Simulation parameters in a package
	struct nfg_SimParams
	{
		double		m_SIGMA_OW_ESTLOC_X;	// Obs noise std dev of pos X 
		double		m_SIGMA_OW_ESTLOC_Y;	// Obs noise std dev of pos Y
		double		m_SIGMA_OW_ESTLOC_Z;	// Obs noise std dev of pos Z
		double		m_SIGMA_OW_ESTVEL_X;	// Obs noise std dev of vel X
		double		m_SIGMA_OW_ESTVEL_Y;	// Obs noise std dev of vel Y
		double		m_SIGMA_OW_ESTVEL_Z;	// Obs noise std dev of vel Z
		double		m_SIGMA_OW_ESTACC_X;	// Obs noise std dev of accel X
		double		m_SIGMA_OW_ESTACC_Y;	// Obs noise std dev of accel Y
		double		m_SIGMA_OW_ESTACC_Z;	// Obs noise std dev of accel Z
		double		m_SIGMA_IN_ESTLOC_X;	// Obs noise std dev of pos X 
		double		m_SIGMA_IN_ESTLOC_Y;	// Obs noise std dev of pos Y
		double		m_SIGMA_IN_ESTLOC_Z;	// Obs noise std dev of pos Z
		double		m_SIGMA_IN_ESTVEL_X;	// Obs noise std dev of vel X
		double		m_SIGMA_IN_ESTVEL_Y;	// Obs noise std dev of vel Y
		double		m_SIGMA_IN_ESTVEL_Z;	// Obs noise std dev of vel Z
		double		m_SIGMA_IN_ESTACC_X;	// Obs noise std dev of accel X
		double		m_SIGMA_IN_ESTACC_Y;	// Obs noise std dev of accel Y
		double		m_SIGMA_IN_ESTACC_Z;	// Obs noise std dev of accel Z
		double		m_BIAS_OW_ESTLOC_X;	// Obs noise bias of pos X 
		double		m_BIAS_OW_ESTLOC_Y;	// Obs noise bias of pos Y
		double		m_BIAS_OW_ESTLOC_Z;	// Obs noise bias of pos Z
		double		m_BIAS_OW_ESTVEL_X;	// Obs noise bias of vel X
		double		m_BIAS_OW_ESTVEL_Y;	// Obs noise bias of vel Y
		double		m_BIAS_OW_ESTVEL_Z;	// Obs noise bias of vel Z
		double		m_BIAS_OW_ESTACC_X;	// Obs noise bias of accel X
		double		m_BIAS_OW_ESTACC_Y;	// Obs noise bias of accel Y
		double		m_BIAS_OW_ESTACC_Z;	// Obs noise bias of accel Z
		double		m_BIAS_IN_ESTLOC_X;	// Obs noise bias of pos X 
		double		m_BIAS_IN_ESTLOC_Y;	// Obs noise bias of pos Y
		double		m_BIAS_IN_ESTLOC_Z;	// Obs noise bias of pos Z
		double		m_BIAS_IN_ESTVEL_X;	// Obs noise bias of vel X
		double		m_BIAS_IN_ESTVEL_Y;	// Obs noise bias of vel Y
		double		m_BIAS_IN_ESTVEL_Z;	// Obs noise bias of vel Z
		double		m_BIAS_IN_ESTACC_X;	// Obs noise bias of accel X
		double		m_BIAS_IN_ESTACC_Y;	// Obs noise bias of accel Y
		double		m_BIAS_IN_ESTACC_Z;	// Obs noise bias of accel Z
		double		m_VIEWTIME;			// Viewing time for S12I to S12
		double		m_TIMESTEPSIZE;		// time resolution for physics model
		double		m_MAXANGLE;			// 1.0 radians, defines movespace of players
		double		m_SIMTIME;			// Simulation time for S34
		double		m_TIME2COLLIDE;		// Approx time to collision
		double		m_ALTITUDE;			// Encounter altitude
		double		m_P1SPEED;			// Speed of aircraft 1
		double		m_P2SPEED;			// Speed of aircraft 2
		double		m_FOV;				// Aircraft 1's total field of view
		double		m_P2ANGLEMINMAX;	// Angle range to draw uniformly over in P2 encounter model
		int			m_NUMSAMPLESPERPROC;	// Number of iid samples of bnet per processor
		int			m_KLEVEL;			// Level-k setting
		int			m_MVALUE;			// number of samples from action distribution (m)
		int			m_MPVALUE;			// number of samples from other's state (m')
		double      m_UTILWGHT;         // amount of weight given to distance to other aircraft (NMAC) in utility function. 
		double		m_BIAS;			// observational bias
	};
	
	/**
	 * Constructor: Construct and set analysis parameters from external
	 *	viewtime - time that players view the state, higher means
	 *		a lower stddev in observations. in seconds.
	 *	timestepsize - resolution of time in seconds for physics sim.
	 *	encounterFile - filename of encounter file to use. Expects an ascii file
	 *		one line per encounter (more lines than encounters requested), 
	 *		2 comma-separated numbers per line, no spaces
	 *		no trailing comma, first number is an integer indicating the configuration
	 *		second number is the angle to offset the initial velocity vector of P2
	 *		e.g.
	 *			**************
	 *			-2,0.05
	 *			1,-0.002
	 *			0,-0.06
	 *			...
	 *			**************
	 */
	net_HitlCollisionAvoid2 ( int argc, char *argv[] );
	~net_HitlCollisionAvoid2 ( void );
	/**
	 * Run: Execute example sequence
	 */
	void			Run ( void );
	
	/**
	 * CPD_S1I_Sample_node, CPD_S1I_Get_probability_prop: Initial
	 *	position of player 1's aircraft
	 */
	class cpd_S1I : public cpd
	{
	public:
		cpd_S1I( const nfg_SimParams& PARAMS ) { m_PARAMS = PARAMS; };
		nfg_Vec		Sample_node ( const nfg_Vecs& paVecs );
		double		Get_probability_prop ( const nfg_Vec& outVec, 
														const nfg_Vecs& paVecs );
	private:
		nfg_SimParams	m_PARAMS;
	};

	/**
	 * CPD_S2I_Sample_node, CPD_S2I_Get_probability_prop: Initial
	 *	position of player 2's aircraft
	 */
	class cpd_S2I : public cpd
	{
	public:
		cpd_S2I( const nfg_SimParams& PARAMS ) { m_PARAMS = PARAMS; };
		nfg_Vec		Sample_node ( const nfg_Vecs& paVecs );
		double		Get_probability_prop ( const nfg_Vec& outVec, 
														const nfg_Vecs& paVecs );
	private:
		nfg_SimParams	m_PARAMS;
	};

	/**
	 * CPD_S12_Sample_node, CPD_S12_Get_probability_prop: Final state of
	 *	the aircraft at the end of the viewtime.  Takes aircraft initial state
	 *	and simulates it forward for viewtime.
	 */
	class cpd_S12 : public cpd
	{
	public:
		cpd_S12( const nfg_SimParams& PARAMS ) { m_PARAMS = PARAMS; };
		nfg_Vec		Sample_node ( const nfg_Vecs& paVecs );
		double		Get_probability_prop ( const nfg_Vec& outVec, 
														const nfg_Vecs& paVecs );
	private:
		nfg_SimParams	m_PARAMS;
	};
	
	/**
	 * CPD_S34_Sample_node, CPD_S34_Get_probability_prop: State of aircraft
	 *	at the end of simulation time.  Takes players' actions and previous
	 *	state of aircraft as input.
	 */
	class cpd_S34 : public cpd
	{
	public:
		cpd_S34( const nfg_SimParams& PARAMS ) { m_PARAMS = PARAMS; };
		nfg_Vec		Sample_node ( const nfg_Vecs& paVecs );
		double		Get_probability_prop ( const nfg_Vec& outVec, 
														const nfg_Vecs& paVecs );
	private:
		nfg_SimParams	m_PARAMS;
	};

	/** 
	 * Util_function_P1: P1's utility function
	 *	instVecs - an instantiation of the net
	 *	returns: utility value
	 */
	double		Util_function_P1 ( const nfg_Inst& instVecs );
		/** 
	 * Util_function_P2: P2's utility function
	 *	instVecs - an instantiation of the net
	 *	returns: utility value
	 */
	double		Util_function_P2 ( const nfg_Inst& instVecs );
	/**
	 * Sample_movespace: Sample a possible move of a decision node 
	 *	(satisficing distribution)
	 *	paVecs - a list of parent vectors
	 *	returns: a realization of the node
	 */
	nfg_Vec		Sample_movespace ( const nfg_Vecs& paVecs );
	/**
	 * Sample_level0: Sample the level 0 distribution of a decision node 
	 *	(level 0 distribution)
	 *	paVecs - a list of parent vectors
	 *	returns: a realization of the node
	 */
	nfg_Vec		Sample_level0 ( const nfg_Vecs& paVecs );
	/**
	 * Analysis1: Performs a series of analyses based on a bnet
	 *	bnet - nfg object
	 */
	void			Analysis1 ( BayesNet& bnet );
	/**
	 * Analysis1_parProc: Support function for Analysis1 that specifies
	 *		the function to be called during parallel processing.  This function will
	 *		be executed on all compute nodes.  This structure is for parallel computing
	 *		capability.
	 *	returns: a list of weighted samples
	 */
	nfg_WtdInsts Analysis1_parProc ( alg_LevelKRelaxed& engine, BayesNet& bnet  );
	/**
	 * Analysis1_postProc: Support function for Analysis1 that specifies
	 *		the function to be called during post processing. This function will only
	 *		execute on the head node.  This structure is for parallel computing
	 *		capability.
	 *	wInsts - a vector of all the results gathered from Analysis2_parProc from all
	 *		compute nodes.
	 */
	void			Analysis1_postProc ( const vector<nfg_WtdInsts>& wInsts  );

	/**
	 * Unit_subtest_file_read: Test code snippet for file read feature. 
	 *		Does not actually test functions from the net_ file, just the logic
	 *		and function calls for file read.
	 */
	static void Unit_subtest_file_read( const string filename );
	

private:
	/// Methods
	/**
	 * Generate_deterministic_S2I: Generate S2I deterministically.
	 *		i.e., same as cpd_S2I::Sample_node but specifying all randomness
	 *	config - an integer that specifies starting position of P2
	 *		thus specifying the encounter configuration. Uniform on [-2, 2].
	 *	phi - a double that is the angle to perturb P2's velocity vector away from 
	 *		the exact collision point. 
	 *		Uniform on [-m_P2ANGLEMINMAX, m_P2ANGLEMINMAX].
	 */
	nfg_Vec Generate_deterministic_S2I ( const int config, const double phi );

	/// Variables
	nfg_SimParams	m_PARAMS;		// Simulation parameters, potentially specified externally
	ifstream		m_encFile;		// Encounter file handle
};

#endif

