
// *************************************************************************
//	LibNFG: Net-form game library
//
//	Author:		Ritchie Lee
//
//  Copyright 2011 United States Government as represented by the Administrator 
//  of the National Aeronautics and Space Administration.  All Rights Reserved.  
//  Licensed under the Apache License, Version 2.0 (the "License"); you may not 
//  use this file except in compliance with the License.  You may obtain a copy 
//  of the License at http://www.apache.org/licenses/LICENSE-2.0
//  Unless required by applicable law or agreed to in writing, software 
//  distributed under the License is distributed on an "AS IS" BASIS, WITHOUT 
//  WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the 
//  License for the specific language governing permissions and limitations 
//  under the License.
//
//	Revision Log
//
//	Username	Date		Description
//	-------------------------------------------------------
//	rlee		20120710	Created
//
// *************************************************************************

#include "net_imergeandspace.h"

#include <boost/lambda/bind.hpp>
#include "ibayesnet.h"
#include "cpd.h"
#include "cpd_uniform.h"
#include "cpd_idecision.h"
#include "cpd_constant.h"
#include "cpd_adder.h"
#include "iplayer.h"
#include "if_bayesnettoolbox.h"

using namespace nfg;
using namespace std;
using namespace boost;
using namespace boost::lambda;

// *************************************************************************
// *************************************************************************
void net_IMergeAndSpace::Run( void )
{
	//*********************************************
	///	 Set up base semi Bayes Net
	
	// Specify vertex names
	string vertexNames[] = 
	{	"T1", "S1", "S2", "D1", "D2", "T2", "S3", "S4"	};

	// Define Edges of the graph
	nfg_Edge edgeArray[] = 
	{ 
		nfg_Edge(S1,D1), nfg_Edge(S1,D2), 
		nfg_Edge(S2,D1), nfg_Edge(S2,D2),
		nfg_Edge(T1,D1), nfg_Edge(T1,D2),
		nfg_Edge(S1,S3), nfg_Edge(D1,S3),
		nfg_Edge(S2,S4), nfg_Edge(D2,S4),
		nfg_Edge(T1,T2)
	};

	// Specify vertex type
	nfg_eVertexType vertexTypes[] =
	{	CHANCE, CHANCE, DECISION, DECISION, CHANCE, CHANCE, CHANCE, CHANCE	};

	// Specify the space of each vertex
	nfg_Spaces spaces(N);
	spaces.at(S1).push_back( nfg_Space1D(1, nfg_Range(0,10.0) ));
	spaces.at(S2).push_back( nfg_Space1D(1, nfg_Range(0,10.0) ));
	spaces.at(D1).push_back( nfg_Space1D(1, nfg_Range(-1.0,1.0) ));
	spaces.at(D2).push_back( nfg_Space1D(1, nfg_Range(-1.0,1.0) ));
	spaces.at(S3).push_back( nfg_Space1D(1, nfg_Range(0,10.0) ));
	spaces.at(S4).push_back( nfg_Space1D(1, nfg_Range(0,10.0) ));

	////	Make base Bayes Net
	// Pass all specified values above to the bnet constructor
	BayesNet baseBnet( "iMnS Base Bnet", N,
		nfg_Edges(edgeArray,edgeArray+sizeof(edgeArray)/sizeof(nfg_Edge)), 
		nfg_Dim(N,1),		// all nodes have size 1
		spaces,
		nfg_Mask(N,true),	// all continuous
		nfg_eVertexTypes(vertexTypes,vertexTypes+N),
		nfg_Mask(N, false),		// no branch nodes
		nfg_Mask(N, false));	// no nodes with dno
	// Set vertex names
	baseBnet.Set_vertex_names ( vector<string>(vertexNames,vertexNames+N) );

	// Base and Kernel bayes net have same 
	//	Declare players
	IPlayer P1 ( bind(&Util_function_P1, _1), bind(&Util_function_P1, _1), "Player1" );	// pass in utility function
	IPlayer P2 ( bind(&Util_function_P2, _1), bind(&Util_function_P2, _1), "Player2" );	// pass in utility function

	//	Create cpds
	// T is time that counts up.  Initialize baseT_cpd to a constant.
	cpd_Constant	baseT_cpd( 0.0 );

	// S1, S2 are uniform on different ranges
	cpd_Uniform		baseS1_cpd ( nfg_Vec(1,4.0), nfg_Vec(1,7.0) );
	cpd_Uniform		baseS2_cpd ( nfg_Vec(1,3.0), nfg_Vec(1,6.0) );
	// Decision nodes
	cpd_IDecision	D1_cpd ( P1 );	// associate with player
	cpd_IDecision	D2_cpd ( P2 );	// associate with player

	// T is time that counts up.  Add a constant
	cpd_Adder	T_cpd( 1.0 );

	// S3, S4 obey aircraft physics
	cpd_S S_cpd;

	//	Link cpds to bnet
	baseBnet.Set_cpd(T1, &baseT_cpd );
	baseBnet.Set_cpd(S1, &baseS1_cpd );
	baseBnet.Set_cpd(S2, &baseS2_cpd );
	baseBnet.Set_cpd(D1, &D1_cpd );
	baseBnet.Set_cpd(D2, &D2_cpd );
	baseBnet.Set_cpd(T2, &T_cpd );
	baseBnet.Set_cpd(S3, &S_cpd );
	baseBnet.Set_cpd(S4, &S_cpd );

	//	Error check
	bassert ( baseBnet.Check() );
	
	//*********************************************
	///	 Set up kernel semi Bayes Net
	
	////	Make kernel Bayes Net
	// Pass all specified values to the bnet constructor
	// In this example, the base and kernel and base semi bayes nets are the same
	// so I am reusing the values as much as possible from above
	BayesNet kernelBnet( "iMnS Kernel Bnet", N,
		nfg_Edges(edgeArray,edgeArray+sizeof(edgeArray)/sizeof(nfg_Edge)), 
		nfg_Dim(N,1),		// all nodes have size 1
		spaces,
		nfg_Mask(N,true),	// all continuous
		nfg_eVertexTypes(vertexTypes,vertexTypes+N),
		nfg_Mask(N, false),		// no branch nodes
		nfg_Mask(N, false));	// no nodes with dno
	// Set vertex names
	kernelBnet.Set_vertex_names ( vector<string>(vertexNames,vertexNames+N) );

	//	Create cpds
	// S's obey aircraft dynamics
	cpd_Glue	glue_cpd;

	//	Link cpds to bnet
	kernelBnet.Set_cpd(T1, &glue_cpd );
	kernelBnet.Set_cpd(S1, &glue_cpd );
	kernelBnet.Set_cpd(S2, &glue_cpd );
	kernelBnet.Set_cpd(D1, &D1_cpd );
	kernelBnet.Set_cpd(D2, &D2_cpd );
	kernelBnet.Set_cpd(T2, &T_cpd );
	kernelBnet.Set_cpd(S3, &S_cpd );
	kernelBnet.Set_cpd(S4, &S_cpd );

	//	Error check
	bassert ( kernelBnet.Check() );

	//*********************************************
	///	 Set up interconnects.  In this example, the 
	/// base and kernel are the same, so the interconnects
	/// will be the same.
	
	// Define Edges of the interconnects
	nfg_Edge icEdgeArray[] = 
	{ 
		nfg_Edge(T2,T1), nfg_Edge(S3,S1), 
		nfg_Edge(S4,S2)
	};

	// Repackage into a vector
	nfg_Edges icEdges(icEdgeArray,icEdgeArray+sizeof(icEdgeArray)/sizeof(nfg_Edge));
	
	//*********************************************
	///	 Set up Iterated semi Bayes Net
	int numSteps = 5;
	IBayesNet ibnet( "Iterated Merging and Spacing", baseBnet,
				kernelBnet, icEdges, icEdges, numSteps );
	

	// Associate players with ibnet
	ibnet.Add_iplayer(P1);
	ibnet.Add_iplayer(P2);

	cout << ibnet.Get_ibnet_name() << endl;
	cout << ibnet.Get_num_steps() << endl;

	//**************************************************
	///	Run main program
	
	// Draw graph (to make sure the code understood your graph correctly)
	if_BayesNetToolbox::Draw_graph( "Output\\iMnS_draw.m", ibnet );
	
	Analysis1(ibnet);

}
// *************************************************************************
// *************************************************************************
nfg_Vec net_IMergeAndSpace::cpd_S::Sample_node ( const nfg_Vecs& paVecs )
{
	// paVecs.at(0) = S;
	// paVecs.at(1) = D;

	double S = paVecs.at(0).at(0);
	double D = paVecs.at(1).at(0);

	double outVal = S+D;				// Add the player's move to the state
	outVal = max( min(outVal,10.0), 0.0 );	// Keep him on the playfield

	// Output as a vector format
	return nfg_Vec(1, outVal);
}
// *************************************************************************
// *************************************************************************
double net_IMergeAndSpace::Util_function_P1 ( const nfg_Inst& instVecs )
{
	//	weights
	double a1 = 0.8;
	double a2 = 1 - a1;

	//	Term 1: How close to the other player, higher is better
	double u1 = abs ( instVecs.at(S2).at(0) - instVecs.at(S1).at(0));

	//	Term 2: How much effort was spent to move, lower is better
	double u2 = abs ( instVecs.at(D1).at(0) );

	// linear combo of all terms
	return a1*u1 - a2*u2;
}
// *************************************************************************
// *************************************************************************
double net_IMergeAndSpace::Util_function_P2 ( const nfg_Inst& instVecs )
{
	// weights
	double a1 = 0.8;
	double a2 = 1 - a1;

	//	Term 1: How close to the other player, higher is better
	double u1 = abs ( instVecs.at(S2).at(0) - instVecs.at(S1).at(0));

	//	Term 2: How much effort was spent to move, lower is better
	double u2 = abs ( instVecs.at(D2).at(0) );

	// linear combo of all terms
	return a1*u1 - a2*u2;
}
// *************************************************************************
// *************************************************************************
nfg_Vec net_IMergeAndSpace::Sample_level0 ( const nfg_Vecs& paVecs )
{
	//	The parent vectors are ignored

	// Go down
	return nfg_Vec(1, -1 );
}
// *************************************************************************
// *************************************************************************
void net_IMergeAndSpace::Analysis1 ( IBayesNet& ibnet )
{
	ibnet.Set_player_policy( 0, bind(&Sample_level0, _1));
	ibnet.Set_player_policy( 1, bind(&Sample_level0, _1));
	
	//nfg_Episode		ep = ibnet.Forward_sample();

	return;
}

