
// *************************************************************************
//	LibNFG: Net-form game library
//
//	Author:		Ritchie Lee
//
//  Copyright 2011 United States Government as represented by the Administrator 
//  of the National Aeronautics and Space Administration.  All Rights Reserved.  
//  Licensed under the Apache License, Version 2.0 (the "License"); you may not 
//  use this file except in compliance with the License.  You may obtain a copy 
//  of the License at http://www.apache.org/licenses/LICENSE-2.0
//  Unless required by applicable law or agreed to in writing, software 
//  distributed under the License is distributed on an "AS IS" BASIS, WITHOUT 
//  WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the 
//  License for the specific language governing permissions and limitations 
//  under the License.
//
//	Revision Log
//
//	Username	Date		Description
//	-------------------------------------------------------
//	rlee		20110721	Created
//
// *************************************************************************

#include "pointmasskinematics.h"

#include <iostream>

using namespace std;

// *************************************************************************
// *************************************************************************
PointMassKinematics::PointMassKinematics( void )
{
	Set_state( 0,0,0,0,0,0 );
}
// *************************************************************************
// *************************************************************************
PointMassKinematics::PointMassKinematics( double x0, double y0, double z0, 
										double vx0, double vy0, double vz0 )
{
	Set_state( x0, y0, z0, vx0, vy0, vz0 );
}
// *************************************************************************
// *************************************************************************
PointMassKinematics::~PointMassKinematics ( void )
{
}
// *************************************************************************
// *************************************************************************
void PointMassKinematics::Set_params ( double dt )
{
	m_dt = dt;
}
// *************************************************************************
// *************************************************************************
void PointMassKinematics::Set_state( double x, double y, double z,
										double vx, double vy, double vz )
{
	// Directly override the current state
	m_x = x;  
	m_y = y;
	m_z = z;
	m_vx = vx;
	m_vy = vy;
	m_vz = vz;
}
// *************************************************************************
// *************************************************************************
void PointMassKinematics::Step ( double ax, double ay, double az )
{
	// Use Euler integration equations of motion
	// for a point mass particle
	m_x = m_x + m_vx * m_dt + ax * m_dt * m_dt / 2;
	m_y = m_y + m_vy * m_dt + ay * m_dt * m_dt / 2;
	m_z = m_z + m_vz * m_dt + az * m_dt * m_dt / 2;
	m_vx = m_vx + ax * m_dt;
	m_vy = m_vy + ay * m_dt;
	m_vz = m_vz + az * m_dt;
}
// *************************************************************************
// *************************************************************************
vector<double> PointMassKinematics::Get_position_3d ( void )
{
	// Load up the data and return
	vector<double> pos3d;
	pos3d.push_back( m_x );
	pos3d.push_back( m_y );
	pos3d.push_back( m_z );

	return pos3d;
}
// *************************************************************************
// *************************************************************************
vector<double> PointMassKinematics::Get_velocity_3d ( void )
{
	// Load up the data and return
	vector<double> vel3d;
	vel3d.push_back( m_vx );
	vel3d.push_back( m_vy );
	vel3d.push_back( m_vz );

	return vel3d;
}
// *************************************************************************
// *************************************************************************
void PointMassKinematics::Unit_test( void )
{
	cout << "Starting PointMassKinematics::Unit_test()..." << endl << endl;

	// Create test object
	PointMassKinematics mass( 0,-10,0, 1,2,3 );

	// time step size parameter
	mass.Set_params( 1.0 );

	// Print out initial positions
	cout << "<" << mass.Get_x() << "," 
					<< mass.Get_y() << ","
					<< mass.Get_z() << ","
					<< mass.Get_vx() << ","
					<< mass.Get_vy() << ","
					<< mass.Get_vz() << ">" << endl;

	// Take some time steps with fixed acceleration and output
	// state to console
	for ( int i = 0; i < 10; i++ )
	{
		mass.Step( 0.0, 0.0, 1.0 );
		cout << "<" << mass.Get_x() << "," 
					<< mass.Get_y() << ","
					<< mass.Get_z() << ","
					<< mass.Get_vx() << ","
					<< mass.Get_vy() << ","
					<< mass.Get_vz() << ">" << endl;
	}

	// Test batch output functions
	vector<double> pos = mass.Get_position_3d();
	vector<double> vel = mass.Get_velocity_3d();

	return;
}


