
// *************************************************************************
//	LibNFG: Net-form game library
//
//	Author:		Ritchie Lee
//
//  Copyright 2011 United States Government as represented by the Administrator 
//  of the National Aeronautics and Space Administration.  All Rights Reserved.  
//  Licensed under the Apache License, Version 2.0 (the "License"); you may not 
//  use this file except in compliance with the License.  You may obtain a copy 
//  of the License at http://www.apache.org/licenses/LICENSE-2.0
//  Unless required by applicable law or agreed to in writing, software 
//  distributed under the License is distributed on an "AS IS" BASIS, WITHOUT 
//  WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the 
//  License for the specific language governing permissions and limitations 
//  under the License.
//
//	Revision Log
//
//	Username	Date		Description
//	-------------------------------------------------------
//	rlee		20110404	Created
//
// *************************************************************************

#include "testcases.h"

#include <iostream>
#include <list>
#include "types.h"

using namespace std;

// *************************************************************************
// *************************************************************************
void TestCases::Add_channel_to_end ( const nfg_tc_Channel& channel, const string name )
{
	m_data.push_back( channel );		// add channel to end of channel list
	m_channelNames.push_back( name );	// add name to the end of channel name list
}
// *************************************************************************
// *************************************************************************
void TestCases::Delete_last_channel ( void )
{
	m_data.pop_back();		// remove last element of channel list
	m_channelNames.pop_back();	// remove last element of name list
}
// *************************************************************************
// *************************************************************************
void TestCases::ComboIterator::operator=( TestCases& src )
{
	//	Set iterator to beginning

	// resize all the indices to the new size
	m_begin.resize( src.Get_num_channels() );
	m_index.resize( src.Get_num_channels() );
	m_end.resize( src.Get_num_channels() );

	// Associate iterator positions for each channel
	int i = 0;
	foreach ( nfg_tc_Channel& ch, src.m_data )
	{
		m_begin.at(i)	= ch.begin();	// start position
		m_index.at(i)	= ch.begin();	// set current position to start
		m_end.at(i)		= ch.end();		// end position
		i++;
	}
}
// *************************************************************************
// *************************************************************************
nfg_tc_TestCase TestCases::ComboIterator::operator()( void )
{
	// To be returned at the end
	nfg_tc_TestCase tcase;

	// if it is "past the end", return an empty testcase
	if ( m_index.empty() )
		return tcase;
	
	// return the current element that index is pointing to for each channel
	foreach( nfg_tc_Channel::iterator& it, m_index )
	{
		tcase.push_back( *it );
	}

	// return the testcase
	return tcase;
}
// *************************************************************************
// *************************************************************************
void TestCases::ComboIterator::operator++( void )
{
	//	Do nothing if it is empty (this marks "past the end" condition)
	if ( m_index.empty() )
		return;

	for ( unsigned int ch = 0; ch < m_index.size(); ch++  )
	{
		//	Try to increment current index (starting from beginning)
		m_index.at(ch)++;
		//	Not at end of a channel
		if ( m_index.at(ch) != m_end.at(ch) )
		{
			break;
		}
		//	Reached the end of a channel, reset this channel
		//	and try to increase the next digit
		else
		{
			m_index.at(ch) = m_begin.at(ch);
			
			//	if this is the last channel, then mark as "past the end"
			if ( ch == m_index.size() - 1 )
			{
				m_index.clear();
				break;
			}
		}
	}
}
// *************************************************************************
// *************************************************************************
void TestCases::ParallelIterator::operator=( TestCases& src )
{
	//	Set iterator to beginning

	// resize all the indices to the new size
	m_begin.resize( src.Get_num_channels() );
	m_index.resize( src.Get_num_channels() );
	m_end.resize( src.Get_num_channels() );

	int i = 0;
	foreach ( nfg_tc_Channel& ch, src.m_data )
	{
		m_begin.at(i)	= ch.begin();	// start position
		m_index.at(i)	= ch.begin();	// set current position to start
		m_end.at(i)		= ch.end();		// end position
		i++;
	}
}
// *************************************************************************
// *************************************************************************
nfg_tc_TestCase TestCases::ParallelIterator::operator()( void )
{
	// To be returned at the end
	nfg_tc_TestCase tcase;

	// if it is "past the end", return an empty testcase
	if ( m_index.empty() )
		return tcase;
	
	// return the current element that index is pointing to for each channel
	foreach( nfg_tc_Channel::iterator& it, m_index )
	{
		tcase.push_back( *it );
	}
	
	// return the testcase
	return tcase;
}
// *************************************************************************
// *************************************************************************
void TestCases::ParallelIterator::operator++( void )
{
	//	Do nothing if it is empty (this marks "past the end" condition)
	if ( m_index.empty() )
		return;

	for ( unsigned int ch = 0; ch < m_index.size(); ch++  )
	{
		//	Try to increment each channel
		m_index.at(ch)++;

		//	If out-of-range, mark as "past end" and return
		if ( m_index.at(ch) == m_end.at(ch) )
		{
			m_index.clear();
			break;
		}
	}
	//	Note that this will stop at the first sign of end for any channel
	//	If the channel are different lengths, it will stop at the end of
	//	the shortest channel
}
// *************************************************************************
// *************************************************************************
void TestCases::Unit_test( void )
{
	// New testcases object
	TestCases tcases;

	// Populate channels (note the different object types)
	nfg_tc_Channel ch1;
	ch1.push_back(1.0); ch1.push_back(2.0); ch1.push_back(3.0);
	
	nfg_tc_Channel ch2;
	ch2.push_back(4); ch2.push_back(5); ch2.push_back(6);

	nfg_tc_Channel ch3;
	ch3.push_back(true); ch3.push_back(false);;

	// Add the channels to the testcases object
	tcases.Add_channel_to_end( ch1, "d" );
	tcases.Add_channel_to_end( ch2, "i" );
	tcases.Add_channel_to_end( ch3, "b" );

	// Test1: combo iterator
	cout << "Test combo iterator:" << endl;
	TestCases::ComboIterator it1;
	for ( it1 = tcases; !it1().empty(); ++it1 )	// iterate through all combinations
	{
		nfg_tc_TestCase tcase = it1();
		cout << tcases.Get_channel_name(0) << "=" << *any_cast<double>( &tcase.at(0) ) << endl;
		cout << tcases.Get_channel_name(1) << "=" << *any_cast<int>( &tcase.at(1) ) << endl;
		cout << tcases.Get_channel_name(2) << "=" << *any_cast<bool>( &tcase.at(2) ) << endl << endl;
	}

	// Test 2: parallel iterator
	cout << "Test parallel iterator: " << endl;
	TestCases::ParallelIterator it2;
	for ( it2 = tcases; !it2().empty(); ++it2 )	// iterate parallel-wise
	{
		nfg_tc_TestCase tcase = it2();
		cout << tcases.Get_channel_name(0) << "=" << *any_cast<double>( &tcase.at(0) ) << endl;
		cout << tcases.Get_channel_name(1) << "=" << *any_cast<int>( &tcase.at(1) ) << endl;
		cout << tcases.Get_channel_name(2) << "=" << *any_cast<bool>( &tcase.at(2) ) << endl << endl;
	}
}

