# -*- coding: utf-8 -*-
"""
Created on Sun Oct 28 10:15:43 2012

@author: jamesbono

A proof of concept script creating a decision node object for the purpose of
converting NFGs into PyMC models after supplying CPTs to decision nodes in an 
outer loop
"""

import pymc as pm
import numpy as np
from types import *
from collections import OrderedDict
from det_chance_node import DetChanceNode

class Decision_Node(object):
    """
    a mockup of an NFG decision node with the following attributes
    ---------
    Arguments
    ---------
    name:           (string - required) name of the decision node, e.g. 'd_12' for player 
                    1's decision in time step 2
    parents:        (dictionary) The values should be NFG discrete node objects, 
                    i.e. 
                    Discr_Chance_Node, 
                    Discr_Deterministic_Node, or 
                    Decision_Node.
                    Default: empty dictionary
                    The key of a parent NFG object should be the name of that
                    parent, e.g. for a parent calle C1, the key should be C1.name 
    space:          (list - required)
    player:         (string) name of the player that owns this decision node, 
                    e.g. '1' for player 1
                    !!!Eventually, this will become a player object!!!
    description:    (string) a verbal description of the decision node, e.g. 
                    'Here player 1 observes x\in {0,1} and y\in[0,3] and 
                    chooses from {a,b,c}'.
                    Default: empty string
    
    -----------------------
    Computed from Arguments
    -----------------------
    CPT:            (numpy array) computed from size of parents and space using
                    the initialize_CPT() method below
    pymc:           (PyMC Stochastic) computed from parents, space and CPT
                    using the convert_to_pymc() method below
    """
    def __init__(self, name, space, player, parents={}, description=""):
        # Checking types of the inputs
        # name is a string
        assert type(name) is StringType, "name is not a string, it is a %r" % type(name)            
        self.name = name
        # player is a string
        assert type(player) is StringType, "player is not a string, it is a %r" % type(player)
        self.player = player
        # description is a string
        assert type(description) is StringType, "description is not a string, it is a %r" % type(description)
        self.description = description
        # parents is a dictionary
        assert type(parents) is DictType, "parents is not a dictionary, it is a %r" % type(parents)
        # fixing the order of the parents dict to be alphabetical by name. This helps
        # with indexing the CPT in the proper order
        self.parents = OrderedDict( sorted(parents.items(), key=lambda t: t[0] ) )
        # space is a list
        assert type(space) is ListType, "space is not a list, it is %r" % type(space)
        self.space = space
        # initializing a pymc attribute (set by convert_to_pymc method below)
        self.pymc = None
                
    def __str__( self ):
        """
        Getting the decision node to reveal basic information about itself.
        """
        return "A decision node called "+self.name
        
    def initialize_CPT( self ):
        """
        Initializing self.CPT from self.parents and self.space. Running this 
        method before setting self.parents and self.space will result in an
        error.
        The indices of self.CPT are [parent1, parent2,..., parentN, self]
        """
        # creating an empty list specifying the size of each dimension of CPT
        CPT_size = []
        # getting the size of the space of each of self's parents.
        """Note: this only allows discrete parents with spaces specified"""
        for par in self.parents:
            CPT_size.append( len(self.parents[par].space) )
        # getting the size of self's space. Note: self is the last dimension
        CPT_size.append( len(self.space) )
        # initializing a CPT of zeros
        self.CPT = np.zeros( CPT_size )
        
    """The ensuing segment of code is a single method"""        
    def convert_to_pymc( self ):
        """
        Converting the NFG node to a PyMC Stochastic object.
        This requires definition of log probability and random functions.
        It returns a PyMC stochastic object.
        """
        
        def node_logp( value, **kwargs ):
            """
            Returns the log probability associated with (value, parents), where
            parents are values passed as *args. Uses self.parents to index into
            self.CPT
            """
            # creating ind to index into the CPT
            ind = []
            for key in self.parents:
                # finding the index of the dim dimension of the CPT                
                ind.append( self.parents[key].space.index( kwargs[key] ) )
            # getting the index of the value
            ind.append( self.space.index( value ) )
            # converting the ind array to a tuple for indexing the CPT
            indo = tuple(ind)
            # getting the log probability of the value, parents index
            p = np.log( self.CPT[indo] )
            return p
            
        def node_rand( **kwargs ):
            """
            Draws a random value from self.space based on the probabilities in
            self.CPT. Uses the parent values in *args to index into a slice of
            self.CPT
            """
            # creating ind to index into the CPT
            ind = []
            for key in self.parents:
                # finding the index of the dim dimension of the CPT                
                ind.append( self.parents[key].space.index( kwargs[key] ) )
            # converting the ind array to a tuple for indexing the CPT
            indo = tuple(ind)
            # getting a cdf from the slice of CPT given by values of parents
            cdf = np.cumsum( self.CPT[indo] )
            # getting a random number
            cutoff = np.random.rand()
            # using cutoff to specify a bin in the cdf and the corresponding
            # value in self.space
            idx = np.nonzero( cdf >= cutoff )[0][0]
            r = self.space[idx]
            return r
        
        # creating an ordered dicitonary from self.parents to preserve
        parents_dict = {}
        for par in self.parents.values():
            # note that parents must have a pymc attribute, even if it's None
            parents_dict[ par.name ] = par.pymc  
        
        # and finally creating the pymc Stochastic object
        self.pymc = pm.Stochastic( logp = node_logp, 
                                    doc = self.description, 
                                    name = self.name, 
                                    parents = parents_dict, 
                                    random = node_rand, 
                                    trace = True, 
                                    value = self.space[0], 
                                    dtype = type( self.space[0] ), 
                                    rseed = 1., 
                                    observed = False, 
                                    cache_depth = 2, 
                                    plot = True, 
                                    verbose = 0 )
        
        # returning the attribute called self.pymc
        return self.pymc
    """This is the end of the convert_to_pymc method"""
